
from ._iwriter import iwriter_base

family = 0
bitplanes = 1

def mantissa(rulestring):

    return {0, 1, 2, 3, 4, 5, 6, 7, 8}

def create_rule(rulestring):

    with open('iterators_gpu.h', 'w') as f:
        f.write('''#define ADVANCE_TILE_64(a, b, tmp, dst) {              \\
    uint64_cu al = ROTL64(a, 1);                       \\
    uint64_cu ar = ROTR64(a, 1);                       \\
    uint64_cu xor2 = al ^ ar;                          \\
    (dst)[threadIdx.x] = xor2 ^ a;                     \\
    __syncthreads();                                   \\
    uint64_cu uda = (dst)[u] & (dst)[d];               \\
    uint64_cu udx = (dst)[u] ^ (dst)[d];               \\
    uint64_cu sv = (xor2 & udx) | uda;                 \\
    uint64_cu pt8 = xor2 ^ udx;                        \\
    uint64_cu sh = al & ar;                            \\
    (tmp)[threadIdx.x] = sh | (a & xor2);              \\
    __syncthreads();                                   \\
    uint64_cu pt4 = ((tmp)[u] ^ (tmp)[d]) ^ (sh ^ sv); \\
    uint64_cu xc3 = ((tmp)[u] | (tmp)[d]) ^ (sh | sv); \\
    b = pt4 & xc3 & (pt8 | a);                         \\\n}\n\n''')
        f.write('#define SKIP_18_GENS 1\n\n')

    with open('iterators_%s.h' % rulestring, 'w') as f:
        f.write('#pragma once\n')
        f.write('#include <stdint.h>\n')
        f.write('#include "../lifeconsts.h"\n')
        f.write('#include "../lifeperm.h"\n')
        f.write('#include "../eors.h"\n')
        f.write('namespace %s {\n\n' % rulestring.replace('-', '_'))

        for iset in [['sse2'], ['sse2', 'avx'], ['sse2', 'avx', 'avx2'],
                        ['sse2', 'avx', 'avx2', 'avx512']]:

            iw = b3s23writer(f, iset)

            if 'avx512' in iset:
                f.write('\n\n#ifdef __AVX512F__\n\n')
                iw.write_monolith(rulestring, 48, 28)
                iw.write_function(rulestring, 48, 28)

            iw.write_function(rulestring, 32, 28)
            iw.write_function(rulestring, 28, 24)
            iw.write_function(rulestring, 24, 20)
            iw.write_function(rulestring, 20, 16)
            iw.write_function(rulestring, 16, 12)
            iw.write_function(rulestring, 12, 8)
            iw.write_iterator()

        f.write('#include "../leaf_iterators_avx512.h"\n')
        f.write('#else\n')
        f.write('#include "../leaf_iterators.h"\n')
        f.write('#endif\n\n')
        f.write('}\n')

class b3s23writer(iwriter_base):

    def load_and_hshift(self, i, oddgen, terminal, from_reg=False, prepare=False):

        d = '(%1)' if (oddgen) else '(%0)'

        if ('avx512' in self.iset):
            n = [64, 48, 32, 16, 64][int(terminal)]
            regname = '%%zmm'
            inreg = regname + str(7 - 7 * (i % 2))
            if from_reg:
                oldreg = regname + str(24 + i)
                self.printinstr('vmovdqa64 %s, %s' % (oldreg, inreg))
            else:
                self.read16n(n, d, 64 * i, 7 - 7 * (i % 2))
        else:
            regbytes = 32 if ('avx2' in self.iset) else 16
            regname = '%%ymm' if (('avx2' in self.iset) and not terminal) else '%%xmm'
            accessor = 'vmovdqu' if ('avx' in self.iset) else 'movups'
            inreg = regname + str(7 - 7 * (i % 2))

            d = d if (i == 0) else (str(regbytes * i) + d)

            self.printinstr('%s %s, %s' % (accessor, d, inreg))

        if prepare:
            self.preparethings(prepare)

        if ('avx' in self.iset):
            self.printinstr('vpsrld $1, %s, %s6' % (inreg, regname))
            self.printinstr('vpslld $1, %s, %s1' % (inreg, regname))
        else:
            self.printinstr('movdqa %s, %s6' % (inreg, regname))
            self.printinstr('movdqa %s, %s1' % (inreg, regname))
            self.printinstr('psrld $1, %s6' % regname)
            self.printinstr('pslld $1, %s1' % regname)

    def horizontal_adders(self, i):

        if ('avx512' in self.iset):
            self.trogicgate(0b11101000, 6, 1, 7 - 7 * (i % 2), 12 - 7 * (i % 2)) # MAJ
            self.logicgate('pxord', 6, 1, 9 - 7 * (i % 2))
            self.logicgate('pxord', 7 - 7 * (i % 2), 9 - 7 * (i % 2), 11 - 7 * (i % 2))
        else:
            self.logicgate('pxor', 6, 1, 9 - 7 * (i % 2))
            self.logicgate('pand', 6, 1, 10 - 7 * (i % 2))
            self.logicgate('pand', 7 - 7 * (i % 2), 9 - 7 * (i % 2), 1)
            self.logicgate('pxor', 7 - 7 * (i % 2), 9 - 7 * (i % 2), 11 - 7 * (i % 2))
            self.logicgate('por', 1, 10 - 7 * (i % 2), 12 - 7 * (i % 2))

    def vertical_bitshifts(self, i):

        if 'avx512' in self.iset:
            self.printinstr('vmovdqa64 %%zmm16, %%zmm18')
            self.printinstr('vmovdqa64 %%zmm17, %%zmm6')
            self.printinstr('vmovdqa64 %%zmm17, %%zmm8')
            self.printinstr('vmovdqa64 %%zmm16, %%zmm20')
            self.printinstr('vmovdqa64 %%zmm16, %%zmm19')
            self.logicgate('permi2d',  9 - 7 * (i % 2), 2 + 7 * (i % 2), 18)
            self.logicgate('permi2d', 11 - 7 * (i % 2), 4 + 7 * (i % 2), 6)
            self.logicgate('permi2d', 12 - 7 * (i % 2), 5 + 7 * (i % 2), 8)
            self.logicgate('permi2d',  7 - 7 * (i % 2), 0 + 7 * (i % 2), 20)
            self.logicgate('permi2d', 12 - 7 * (i % 2), 5 + 7 * (i % 2), 19)
        elif 'avx2' in self.iset:
            self.logicgate('pblendd $1,',  9 - 7 * (i % 2), 2 + 7 * (i % 2), 2 + 7 * (i % 2))
            self.logicgate('pblendd $1,', 10 - 7 * (i % 2), 3 + 7 * (i % 2), 3 + 7 * (i % 2))
            self.logicgate('pblendd $3,', 11 - 7 * (i % 2), 4 + 7 * (i % 2), 6)
            self.logicgate('pblendd $3,', 12 - 7 * (i % 2), 5 + 7 * (i % 2), 8)
            self.logicgate('pblendd $1,',  7 - 7 * (i % 2), 0 + 7 * (i % 2), 0 + 7 * (i % 2))
            self.logicgate('permd', 2 + 7 * (i % 2), 13, 2 + 7 * (i % 2))
            self.logicgate('permd', 3 + 7 * (i % 2), 13, 3 + 7 * (i % 2))
            self.printinstr('vpermq $57, %%ymm6, %%ymm6')
            self.printinstr('vpermq $57, %%ymm8, %%ymm8')
            self.logicgate('permd', 0 + 7 * (i % 2), 13, 0 + 7 * (i % 2))
        else:
            self.logicgate('pand', 13, 2 + 7 * (i % 2), 2 + 7 * (i % 2))
            self.logicgate('pand', 13, 3 + 7 * (i % 2), 3 + 7 * (i % 2))
            self.logicgate('pand', 13, 0 + 7 * (i % 2), 0 + 7 * (i % 2))
            self.logicgate('pandn',  9 - 7 * (i % 2), 13, 1)
            self.logicgate('por', 1, 2 + 7 * (i % 2), 2 + 7 * (i % 2))
            self.logicgate('pandn', 10 - 7 * (i % 2), 13, 1)
            self.logicgate('por', 1, 3 + 7 * (i % 2), 3 + 7 * (i % 2))
            self.logicgate('pandn',  7 - 7 * (i % 2), 13, 1)
            self.logicgate('por', 1, 0 + 7 * (i % 2), 0 + 7 * (i % 2))
            self.logicgate('shufps $0x39,',  2 + 7 * (i % 2), 2 + 7 * (i % 2), 2 + 7 * (i % 2))
            self.logicgate('shufps $0x39,',  3 + 7 * (i % 2), 3 + 7 * (i % 2), 3 + 7 * (i % 2))
            self.logicgate('shufps $0x4e,', 11 - 7 * (i % 2), 4 + 7 * (i % 2), 6)
            self.logicgate('shufps $0x4e,', 12 - 7 * (i % 2), 5 + 7 * (i % 2), 8)
            self.logicgate('shufps $0x39,',  0 + 7 * (i % 2), 0 + 7 * (i % 2), 0 + 7 * (i % 2))

    def everything_else(self, i):
        '''
        This routine ensures that the next cell state resides in register 6.
        '''

        if 'avx512' in self.iset:
            self.trogicgate(0b10010110, 18, 4 + 7 * (i % 2), 6) # XOR8
            self.trogicgate(0b10110010, 18, 6, 4 + 7 * (i % 2)) # MAJ(XOR2, XOR3, XOR3)
            self.trogicgate(0b10100110, 4 + 7 * (i % 2), 19, 18) # XOR(g, AND(19, NOT(18)))
            self.trogicgate(0b00010110, 18, 5 + 7 * (i % 2), 8) # 1OF3
            self.trogicgate(0b10100010, 8, 4 + 7 * (i % 2), 18) # AND(f, OR(i, NOT(g)))
            self.trogicgate(0b10101000, 18, 20, 6) # AND(h, OR(d, 20))
        else:
            self.logicgate('pxor', 4 + 7 * (i % 2), 6, 6)
            self.logicgate('pxor', 6, 2 + 7 * (i % 2), 2 + 7 * (i % 2))
            self.logicgate('por', 6, 4 + 7 * (i % 2), 1)
            self.logicgate('pand', 2 + 7 * (i % 2), 6, 6)
            self.logicgate('pandn', 1, 6, 6)

            self.logicgate('por', 2 + 7 * (i % 2), 0 + 7 * (i % 2), 0 + 7 * (i % 2)) # odd | centre-on

            self.logicgate('pxor', 5 + 7 * (i % 2), 6, 1)
            self.logicgate('por',  5 + 7 * (i % 2), 6, 6)
            self.logicgate('por',  3 + 7 * (i % 2), 8, 2 + 7 * (i % 2))
            self.logicgate('pxor', 3 + 7 * (i % 2), 8, 8)
            self.logicgate('pxor', 2 + 7 * (i % 2), 6, 6)
            self.logicgate('pxor', 8, 1, 1)
            self.logicgate('pand', 1, 6, 6)
            self.logicgate('pand', 0 + 7 * (i % 2), 6, 6)


    def save_result_avx512(self, i, oddgen, terminal, diff=False, to_reg=False):

        n = [64, 48, 32, 16, 64][int(terminal)]

        e = '(%0)' if oddgen else '(%1)'
        offset = 64 * (i - 1) + (8 if oddgen else 0)

        if diff:
            self.read16n(n, e, offset, 8)
            self.trogicgate(0b11100100, 14, 8, 6) # Use reg14 as a mask
            self.write16n(n, 6, offset, e)
            if (diff == 'initial'):
                self.logicgate('pxord', 6, 8, 15)
                self.write16n(64, 15, 0, '(%1)')
            elif (diff == 'final'):
                self.logicgate('pxord', 6, 8, 8)
                if (n == 48):
                    self.printinstr('vshufi32x4 $78, %s, %s, %s' % ('%%zmm15', '%%zmm15', '%%zmm14'))
                    self.printinstr('vshufi32x4 $78, %s, %s, %s' % ('%%zmm8', '%%zmm8', '%%zmm13'))
                    self.printinstr('vpternlogd $254, %%zmm8, %%zmm14, %%zmm15')
                    self.write16n(16, 13, 96, '(%1)')
                    self.write16n(32, 15, 64, '(%1)')
                else:
                    self.printinstr('vshufi32x4 $78, %s, %s, %s' % ('%%zmm15', '%%zmm15', '%%zmm13'))
                    self.logicgate('pord', 13, 15, 15)
                    self.write16n( n,  8, 64, '(%1)')
                    self.write16n(32, 15, 32, '(%1)')
            else:
                self.trogicgate(0b11110110, 6, 8, 15)
        elif to_reg:
            oldreg = '%%zmm' + str(23 + i)
            self.printinstr('vmovdqa64 %s, %s' % ('%%zmm6', oldreg))
        else:
            self.write16n(n, 6, offset, e)

    def save_result(self, i, oddgen, terminal, diff=False):

        regbytes = 32 if ('avx2' in self.iset) else 16
        if oddgen:
            e = str(regbytes * (i - 1) + 8) + '(%0)'
        else:
            e = '(%1)' if (i == 1) else (str(regbytes * (i - 1)) + '(%1)')

        regname = '%%ymm' if (('avx2' in self.iset) and not terminal) else '%%xmm'
        accessor = 'vmovdqu' if ('avx' in self.iset) else 'movups'
        if diff:
            self.logicgate('pand', 14, 6, 6, regname)
            self.printinstr('%s %s, %s8' % (accessor, e, regname))
            self.logicgate('pandn', 8, 14, 1, regname)
            self.logicgate('por', 6, 1, 1, regname)
            self.printinstr('%s %s1, %s' % (accessor, regname, e))
            regname = '%%ymm' if ('avx2' in self.iset) else '%%xmm'
            if (diff == 'initial'):
                self.logicgate('pxor', 1, 8, 15)
                self.printinstr('%s %s15, %s' % (accessor, regname, '(%1)'))
            else:
                self.logicgate('pxor', 1, 8, 8)
                self.logicgate('por', 8, 15, 15)
            if (diff == 'final'):
                pos2 = '64(%1)' if ('avx2' in self.iset) else '32(%1)'
                pos1 = '32(%1)' if ('avx2' in self.iset) else '16(%1)'
                self.printinstr('%s %s8, %s' % (accessor, regname, pos2))
                self.printinstr('%s %s15, %s' % (accessor, regname, pos1))
        else:
            self.printinstr('%s %s6, %s' % (accessor, regname, e))

    def prologue(self):

        self.f.write('        asm (\n')

    def epilogue(self, dwidth):

        self.f.write('                : /* no output operands */ \n')
        self.f.write('                : "r" (d), "r" (e)')
        if (dwidth):
            self.f.write(', "r" (apg::__sixteen%d)' % dwidth)
        self.f.write('\n')
        self.f.write('                : "ebx", ')
        for i in range(21 if ('avx512' in self.iset) else 16):
            self.f.write('"xmm%d", ' % i)
            if (i % 6 == 4):
                self.f.write('\n' + (' ' * 20))
        self.f.write('"memory");\n\n')

    def assemble(self, rulestring, oddgen, rowcount, dwidth, regtemp=False):

        prepare = False
        if (not regtemp) or (oddgen == 0):
            self.prologue()
            prepare = dwidth

        rpr = 16 if ('avx512' in self.iset) else (8 if ('avx2' in self.iset) else 4)

        iters  = (rowcount - (8 if oddgen else 4)) // rpr + 2
        riters = (rowcount - 4) // rpr + 1

        for i in range(iters):

            if (i < riters):
                terminal = max(0, ((i + 1) * rpr - rowcount) // 4)
                self.load_and_hshift(i, oddgen, terminal, from_reg=(regtemp and oddgen), prepare=prepare)
                prepare = False
                self.horizontal_adders(i)

            if (i > 0):
                self.vertical_bitshifts(i)
                self.everything_else(i)
                terminal = max(0, (i * rpr + (4 if oddgen else 0) - rowcount) // 4)

                if oddgen:
                    if (i == 1):
                        diff = 'initial'
                    elif (i + 1 == iters):
                        diff = 'final'
                    else:
                        diff = True
                else:
                    diff = False

                if 'avx512' in self.iset:
                    self.save_result_avx512(i, oddgen, terminal, diff, to_reg=(regtemp and not oddgen))
                else:
                    self.save_result(i, oddgen, terminal, diff)

        if (not regtemp) or (oddgen == 1):
            self.epilogue(dwidth)

    def write_monolith(self, rulestring, rowcount, dwidth):

        name = 'iterate_%s_%d_%d_monolith' % (self.besti, rowcount, dwidth)

        params = 'uint32_t * __restrict__ diffs'
        for i in 'ed':
            params = 'uint32_t * __restrict__ ' + i + ', ' + params

        self.f.write('    bool %s(%s) {\n' % (name, params))

        self.assemble(rulestring, 0, rowcount, dwidth, regtemp=True)
        self.assemble(rulestring, 1, rowcount, dwidth, regtemp=True)

        if 'avx512' in self.iset:
            newrows = rowcount - 4

            while (newrows >= 32):
                newrows -= 16

            if (newrows <= 16):
                bindices = list(range(0, newrows // 2))
            elif (newrows == 28):
                bindices = [8, 9, 10, 11, 12, 13]
            else:
                bindices = list(range(4, newrows // 2))
            lindices = list(range(newrows - 2, newrows))

        elif 'avx2' in self.iset:

            bindices = [4, 5, 6, 7]
            lindices = list(range(16 + ((rowcount - 6) % 8), 18 + ((rowcount - 6) % 8)))

        else:

            bindices = [2, 3]
            lindices = list(range(8 + ((rowcount - 6) % 4), 10 + ((rowcount - 6) % 4)))

        self.f.write('        uint64_t* e64 = ((uint64_t*) e);\n')
        self.f.write('        uint64_t bigdiff = %s;\n' % (' | '.join(['e64[%d]' % x for x in bindices])))
        self.f.write('        if (diffs != 0) {\n')
        self.f.write('            diffs[0] = (bigdiff | (bigdiff >> 32));\n')
        self.f.write('            diffs[1] = e[0] | e[1];\n')
        self.f.write('            diffs[2] = %s;\n' % (' | '.join(['e[%d]' % x for x in lindices])))
        self.f.write('        }\n')
        self.f.write('        return (bigdiff == 0);\n')
        self.f.write('    }\n\n')



'''
The following code was written by M. I. Wright (for Python 3) and
subsequently modified to be compatible with both versions of Python.
'''

from ._wright import *

ACTIVE, INACTIVE = 'active', 'inactive'

def unbind_vars(transition, start=0):
    """
    Suffix varnames in transition with locally-unique numbers so that Golly
    doesn't bind the identical names
    """
    ret, seen = [], {}
    for state in transition:
        if isinstance(state, int) or state.isdigit():
            ret.append(state)
        else:
            seen[state] = cur = seen.get(state, -1) + 1
            ret.append('{}_{}'.format(state, cur))
    return ret


def _lazy_tr(states):
    """
    Take sequence of 2- or 3-tuples, interpreting them as
    (
      cellstate value to repeat+unbind,
      number of times to repeat,
      optional value to start bindings at,
    )
    and produce an expanded/chained form thereof
    """
    for value in states:
        if isinstance(value, tuple):
            state = value[0]
            count = value[1]
            start_at = value[2:]
            for uv in unbind_vars([state] * count, *start_at):
                yield uv
        else:
            yield value


def tr(*states):
    """Generate a transition (as list) from varargs"""
    return list(_lazy_tr(states))


def make_totalistic(birth, survival):
    transitions = []
    # Birth
    transitions += [tr(0, (ACTIVE, n), (INACTIVE, 8 - n), 1) for n in map(int, birth)]
    # Survival
    transitions += [tr((ACTIVE, n + 1), (INACTIVE, 8 - n), ACTIVE+'_0') for n in map(int, survival)]
    return 'permute', transitions


def make_nontot(birth, survival):
    transitions = []
    birth, survival = combine_rstring(birth), combine_rstring(survival)
    # Birth
    transitions += [unbind_vars((0,) + tuple(([INACTIVE, ACTIVE][int(c)] for c in NAPKINS[total][configuration])) + (1,))
        for total, configurations in birth.items()
        for configuration in configurations]
    # Survival
    transitions += [unbind_vars((ACTIVE,) + tuple(([INACTIVE, ACTIVE][int(c)] for c in NAPKINS[total][configuration]))) + [ACTIVE+'_0']
        for total, configurations in survival.items()
        for configuration in configurations]
    return 'rotate4reflect', transitions


def make_rule(birth, survival, age_pattern):
    active, inactive = [], [0]
    n_states = 1 + sum(age_pattern)
    tr_func = make_totalistic if (birth + survival).isdigit() else make_nontot
    
    if n_states > 255:
        raise ValueError("State count %d exceeds maximum of 255." % n_states)
    
    symmetry_type, transitions = tr_func(birth, survival)

    # Transition toward death where unspecified
    transitions += [tr(state, ('all', 8), (state + 1) % n_states) for state in range(1, n_states)]  # state 0 isn't included
    
    lower = 1  # state 0 is always 'inactive', so we start at 1
    for idx, upper in enumerate(age_pattern):
        if (idx % 2):
            inactive += list(range(lower, lower + upper))
        else:
            active += list(range(lower, lower + upper))
        lower += upper
    return symmetry_type, n_states, active, inactive, transitions


def write_table(fp, rulename, symmetries, n_states, active, inactive, transitions):
    fp.write('@RULE {}\n@TABLE\n'.format(rulename))
    fp.write('n_states:{}\nneighborhood:Moore\nsymmetries:{}\n'.format(n_states, symmetries))
    
    # Variables
    def define_var(name, var):
        if not var:
            raise SystemExit(
              'ERROR: Var for {!r} states is coming up empty. '
              'Are you sure this is a valid rule?'.format(name)
              )
        fp.write('\nvar {0}_0={1}\n'.format(name, list(var)).replace('[', '{').replace(']', '}'))
        for n in range(8):
            fp.write('var {0}_{1}={0}_0\n'.format(name, n + 1))

    define_var(ACTIVE, active)
    define_var(INACTIVE, inactive)
    define_var('all', range(n_states))

    # Transitions
    for tr in transitions:
        fp.write('\n' + ','.join(map(str, tr)))


def create_rule(rulestring):

    transitions = []
    bs, age_pattern = rulestring.split('d')
    age_pattern = tuple(map(int, age_pattern.split('-')))

    str2tab(bs) # Raise a NonCanonicalError if initial segment is non-canonical

    birth, survival = tuple(map(str.strip, bs[1:].split('s')))

    ruletable = make_rule(birth, survival, age_pattern)

    # Write temporary .rule file:
    with open((rulestring+'.rule'), 'w') as fp:
        write_table(fp, rulestring, *ruletable)

    # Convert .rule file into C/C++ code:
    rule2files(rulestring + '.rule')

    # Link code into lifelib:
    make_header(rulestring)



print('switch (sqt->updateflags & 63) {')

for i in list(range(64))[::-1]:

    print('case %d:' % i)

    if (i & 1):
        print('    sqt->copyBoundary0(sqt->neighbours[0]);')
        continue

    if (i & 8):
        print('    sqt->copyBoundary3(sqt->neighbours[3]);')

    if (i & 6 == 6):
        print('    sqt->copyBoundary12(sqt->neighbours[1], sqt->neighbours[2]);')
    if (i & 6 == 2):
        print('    sqt->copyBoundary1(sqt->neighbours[1]);')
    if (i & 6 == 4):
        print('    sqt->copyBoundary2(sqt->neighbours[2]);')

    if (i & 48 == 48):
        print('    sqt->copyBoundary45(sqt->neighbours[4], sqt->neighbours[5]);')
    if (i & 48 == 16):
        print('    sqt->copyBoundary4(sqt->neighbours[4]);')
    if (i & 48 == 32):
        print('    sqt->copyBoundary5(sqt->neighbours[5]);')

    print('    break;')

print('}')





import unittest
import lifelib.genera.rulefiles.parsetable as parsetable

from lifelib.genera import sanirule

class TestPerm(unittest.TestCase):

    def test_permute(self):

        generators = parsetable.permute_symmetry(parsetable.nhoods['Moore'])
        self.assertEqual(generators, [[(1, 2, 3, 4, 5, 6, 7, 8)], [(1, 2)]])

    def test_sanirule(self):

        self.assertEqual(sanirule('B3/S23'), 'b3s23')

if __name__ == '__main__':
    unittest.main()



from .parsetable import GetNumberOfInputs
from ast import literal_eval

from .writetree import RuleTree

from .automorph import matperms
from ..isotropic import tab2str, str2tab
from ..exceptions import SurplusTreeError

# Named neighbourhoods:
nhoods = {'2': [(-1, 0), (1, 0), (0, 0), (0, 0)],
          '4': [(0, -1), (-1, 0), (1, 0), (0, 1), (0, 0), (0, 0)],
          '8': [(-1, -1), (1, -1), (-1, 1), (1, 1), (0, -1), (-1, 0), (1, 0), (0, 1), (0, 0), (0, 0)]}

def ParseRuleTree(list_of_lines):

    numInputs = 0
    num_nodes = 0
    n_states = 0

    list_of_nodes = []

    for line in list_of_lines:

        line = line.split('#')[0].strip()

        if (line == '') or (line[0] == '@'):
            continue

        if '=' in line:

            command, thing = tuple([x.strip() for x in line.split('=')])

            if command == 'num_states':
                n_states = int(thing)
                if (n_states < 2) or (n_states > 65536):
                    raise ValueError('n_states must be between 2 and 65536, inclusive')
            elif command in ['num_neighbors', 'num_neighbours']:
                nstring = thing
                if nstring in nhoods:
                    nhood = nhoods[nstring]
                else:
                    nhood = [tuple(x) for x in literal_eval(nstring)]
                numInputs = GetNumberOfInputs(nhood)
            elif command == 'num_nodes':
                num_nodes = int(thing)

        else:

            # Replace all non-digits with whitespace:
            line = ''.join([(c if (c in '0123456789') else ' ') for c in line]).strip()

            if line == '':
                continue

            if (n_states <= 0) or (num_nodes <= 0) or (numInputs <= 0):
                raise RuntimeError('num_states, num_neighbours, and num_nodes must all be defined prior to the first node.')

            # Tree node
            numbers = [int(x) for x in line.split()]
            list_of_nodes.append(numbers)

    if (len(list_of_nodes) != num_nodes):
        raise RuntimeError('%d nodes found; %d expected' % (len(list_of_nodes), num_nodes))

    return n_states, nhood, list_of_nodes


def compress_tree(list_of_nodes, n_states, numInputs, seqonly=True):
    rt = RuleTree(n_states, numInputs, initialise=False)
    rt.seq = list(map(tuple, list_of_nodes))
    rt._shrink()
    if seqonly:
        return list(map(list, rt.seq))
    else:
        return rt


def make_tree(n_states, nhood, list_of_nodes):

    numInputs = GetNumberOfInputs(nhood)
    return compress_tree(list_of_nodes, n_states, numInputs, seqonly=False)


def normalise_neighbourhood(nhood):

    x0, y0 = nhood[-1]
    return [(x - x0, y - y0) for (x, y) in nhood]


def get_symmetries(n_states, nhood, list_of_nodes):

    numInputs = GetNumberOfInputs(nhood)
    rt = compress_tree(list_of_nodes, n_states, numInputs, seqonly=False)

    pts = nhood[:numInputs]
    perms = matperms(pts)

    perms = [p for p in perms if rt.invariant(p)]
    return perms


def optimise_tree(n_states, nhood, list_of_nodes):

    numInputs = GetNumberOfInputs(nhood)

    # Perform optimisation of this tree:
    list_of_nodes = compress_tree(list_of_nodes, n_states, numInputs)

    # Determine which inputs are essential:
    essentials = [False] * numInputs
    for node in list_of_nodes:
        if len(set(node[1:])) > 1:
            essentials[-node[0]] = True

    # Flatten unnecessary layers:
    for node in list_of_nodes:
        depth = node[0]
        if (depth > 1):
            next_depth = list_of_nodes[node[-1]][0]
            if not essentials[-next_depth]:
                for j in range(1, len(node)):
                    node[j] = list_of_nodes[node[j]][-1]
    while not essentials[-list_of_nodes[-1][0]]:
        x = list_of_nodes[-1][-1]
        list_of_nodes = list_of_nodes[:x+1]

    # Rename inputs:
    emap = [0] * (numInputs + 1)
    for i in range(1, numInputs + 1):
        emap[i] = emap[i-1] + (1 if essentials[-i] else 0)
    for node in list_of_nodes:
        node[0] = max(1, emap[node[0]])

    # Shrink neighbourhood:
    new_inputs = [t for (x, t) in zip(essentials, nhood[:numInputs]) if x]
    nhood = new_inputs + nhood[numInputs:]
    numInputs = len(new_inputs)

    # Reoptimise:
    list_of_nodes = compress_tree(list_of_nodes, n_states, numInputs)

    if len(nhood) == numInputs + 1:
        nhood = normalise_neighbourhood(nhood)

    return n_states, nhood, list_of_nodes


def TreeToArray(nhood, list_of_nodes):

    inputs = nhood[:GetNumberOfInputs(nhood)]

    lord2 = []

    indices =  {(-1, -1): 0, (0, -1): 1, (1, -1): 2,
                (-1,  0): 3, (0,  0): 4, (1,  0): 5,
                (-1,  1): 6, (0,  1): 7, (1,  1): 8}

    for i in range(512):

        p = len(list_of_nodes) - 1

        for k in inputs:
            s = (i >> indices[k]) & 1
            p = list_of_nodes[p][s+1]

        lord2.append(p)

    return lord2


def ReadRuleTree(list_of_lines):

    # Parse the actual file into a rule tree:
    ruletree = ParseRuleTree(list_of_lines)

    # Optimise the resulting rule tree:
    return optimise_tree(*ruletree)


def FlattenRuleTree(n_states, nhood, list_of_nodes):

    # Determine whether rule is 2-state isotropic:
    if (n_states == 2) and (GetNumberOfInputs(nhood) + 1 == len(nhood)):
        chebyshev_range = max([abs(x) for y in nhood for x in y])
        if (chebyshev_range <= 1):
            rarray = TreeToArray(nhood, list_of_nodes)

            for hexagonal in [False, True]:
                rstring = tab2str(rarray, hexagonal)
                if (rarray == str2tab(rstring)):
                    raise SurplusTreeError(rstring)

    # Serialise tree to an array:
    num_nodes = len(list_of_nodes)

    def serialise_node(numbers):

        depth = numbers[0]
        children = numbers[1:]

        if depth > 1:
            # Convert node number to index into uint32_t array:
            children = [n_states * (num_nodes - 1 - x) for x in children]

        return children

    return [serialise_node(data) for data in list_of_nodes[::-1]]



from ast import literal_eval

# Named neighbourhoods:
nhoods = {'Moore': [(0, 0), (0, -1), (1, -1), (1, 0), (1, 1), (0, 1), (-1, 1), (-1, 0), (-1, -1), (0, 0)],
          'vonNeumann': [(0, 0), (0, -1), (1, 0), (0, 1), (-1, 0), (0, 0)],
          'hexagonal': [(0, 0), (0, -1), (1, 0), (1, 1), (0, 1), (-1, 0), (-1, -1), (0, 0)],
          'oneDimensional': [(0, 0), (-1, 0), (1, 0), (0, 0)]}

# Named symmetries. Each symmetry group is expressed as a list of generators,
# where each generator is expressed as a list of disjoint cycles:
syms = {('Moore', 'rotate2'):                 [[(1, 5), (2, 6), (3, 7), (4, 8)]],
        ('Moore', 'rotate2reflect'):          [[(1, 5), (2, 6), (3, 7), (4, 8)], [(2, 8), (3, 7), (4, 6)]],
        ('Moore', 'rotate4'):                 [[(1, 3, 5, 7), (2, 4, 6, 8)]],
        ('Moore', 'rotate4reflect'):          [[(1, 3, 5, 7), (2, 4, 6, 8)], [(2, 8), (3, 7), (4, 6)]],
        ('Moore', 'rotate8'):                 [[(1, 2, 3, 4, 5, 6, 7, 8)]],
        ('Moore', 'rotate8reflect'):          [[(1, 2, 3, 4, 5, 6, 7, 8)], [(2, 8), (3, 7), (4, 6)]],
        ('Moore', 'reflect_horizontal'):      [[(2, 8), (3, 7), (4, 6)]],

        ('vonNeumann', 'rotate2'):            [[(1, 3), (2, 4)]],
        ('vonNeumann', 'rotate2reflect'):     [[(1, 3)], [(2, 4)]],
        ('vonNeumann', 'rotate4'):            [[(1, 2, 3, 4)]],
        ('vonNeumann', 'rotate4reflect'):     [[(1, 2, 3, 4)], [(2, 4)]],
        ('vonNeumann', 'reflect_horizontal'): [[(2, 4)]],

        ('hexagonal', 'rotate2'):             [[(1, 4), (2, 5), (3, 6)]],
        ('hexagonal', 'rotate3'):             [[(1, 3, 5), (2, 4, 6)]],
        ('hexagonal', 'rotate6'):             [[(1, 2, 3, 4, 5, 6)]],
        ('hexagonal', 'rotate6reflect'):      [[(1, 2, 3, 4, 5, 6)], [(1, 6), (2, 5), (3, 4)]],
        ('hexagonal', 'reflect_horizontal'):  [[(1, 6), (2, 5), (3, 4)]],

        ('oneDimensional', 'reflect'):        [[(1, 2)]]}

def cycles2perm(n, g):
    '''
    Idempotent function for converting a list of disjoint cycles
    into a permutation.
    '''

    if len(g) == 0:
        return tuple(range(n))
    elif hasattr(g[0], '__iter__'):
        perm = list(range(n))
        for cyc in g:
            for i in range(len(cyc)):
                perm[cyc[i-1]] = cyc[i]
        return tuple(perm)
    elif set(g) == set(range(n)):
        return tuple(g)
    else:
        raise ValueError("%s is not in S_%d" % (g, n))

def permute_by_generators(generators, initial, used=set([])):
    '''
    Produce the subgroup of S_n generated by the generators.
    '''

    initial = tuple(initial) # ensure hashable

    if initial in used:
        return []

    pg = [initial]
    used.add(initial)

    # The next four lines allow us to iterate over a list whilst
    # appending to it. The more Pythonic 'for h in pg' is not
    # guaranteed by the standard to work as intended.
    i = 0
    while (i < len(pg)):
        h = pg[i]
        i += 1

        # Compose h with each of our generators and include any
        # previously-unseen elements to the permutation group:
        for g in generators:
            f = tuple([h[x] for x in g]) # Compose permutations
            if f not in used:
                used.add(f)
                pg.append(f)

    # We now have the desired subgroup:
    return pg

def GetNumberOfInputs(nhood):

    # Find which elements are inputs and which are outputs:
    numInputs = len(nhood) - 1
    for i in range(len(nhood)):
        if nhood[i] in nhood[:i]:
            numInputs = i
            break

    return numInputs

def permute_symmetry(nhood):
    '''
    Produces generators for the 'permute' symmetry group:
    '''

    numInputs = GetNumberOfInputs(nhood)
    indices = [[], [], [], []]
    symmetry = []

    for (i, j) in enumerate(nhood):
        if j in nhood[i+1:]:
            indices[2].append(i)
            indices[3].append(i+1+nhood[i+1:].index(j))
        elif i < numInputs:
            indices[0].append(i)
        else:
            indices[1].append(i)

    if len(indices[0]) >= 2:
        symmetry.append([tuple(indices[0])])
    if len(indices[0]) >= 3:
        symmetry.append([tuple(indices[0][:2])])
    if len(indices[1]) >= 2:
        symmetry.append([tuple(indices[1])])
    if len(indices[1]) >= 3:
        symmetry.append([tuple(indices[1][:2])])
    if len(indices[2]) >= 2:
        symmetry.append([tuple(indices[2]), tuple(indices[3])])
    if len(indices[2]) >= 3:
        symmetry.append([tuple(indices[2][:2]), tuple(indices[3][:2])])

    return symmetry


def read_transition(line, variables, transitions, used, numParams, symmetry):
    '''
    Subroutine of ReadRuleTable()
    '''

    if ',' in line:
        entries = line.replace('\n','').replace(',',' ').replace(';',' ').replace(':',' ').split()
    else:
        entries = list(line.strip()) # special no-comma format

    if len(entries) != numParams:
        raise RuntimeError('Wrong number of entries on line: '+line+' (expected '+str(numParams)+')')

    # retrieve the variables that repeat within the transition, these are 'bound'
    bound_vars = [ e for e in set(entries) if entries.count(e)>1 and e in variables ]

    # iterate through all the possible values of each bound variable
    var_val_indices = {k : 0 for k in bound_vars}

    while True:

        def f(e):
            if e in bound_vars:
                return [variables[e][var_val_indices[e]]]
            elif e in variables:
                return variables[e]
            else:
                return [int(e)]

        transition = [tuple(f(e)) for e in entries]

        # expand for symmetry using the explicit list
        pg = permute_by_generators(symmetry, transition, used)
        transitions += [list(map(list, x)) for x in pg]

        # increment the variable values (or break out if done)
        var_val_to_change = 0
        while var_val_to_change<len(bound_vars):
            var_label = bound_vars[var_val_to_change]
            if var_val_indices[var_label]<len(variables[var_label])-1:
                var_val_indices[var_label] += 1
                break
            else:
                var_val_indices[var_label] = 0
                var_val_to_change += 1
        if var_val_to_change >= len(bound_vars):
            break

def ReadRuleTable(list_of_lines):
    '''
    Based on the ReadRuleTable code from Golly's glife module, largely
    written by Tim Hutton, Tom Rokicki, and Andrew Trevorrow.
    '''

    variables   = {}
    transitions = []
    used = set([])

    for line in list_of_lines:

        line = line.split('#')[0].strip()

        if (line == '') or (line[0] == '@'):
            continue

        if ':' in line:

            command, thing = tuple([x.strip() for x in line.split(':')])

            if command in ['n_states', 'states']:
                n_states = int(thing)
                if (n_states < 2) or (n_states > 65536):
                    raise ValueError('n_states must be between 2 and 65536, inclusive')
            elif command in ['neighbourhood', 'neighborhood']:
                nstring = thing
                if nstring in nhoods:
                    nhood = nhoods[nstring]
                else:
                    nhood = [tuple(x) for x in literal_eval(nstring)]
                numParams = len(nhood)
            elif command == 'symmetries':
                symstring = thing
                if (nstring, symstring) in syms:
                    symmetry = syms[nstring, symstring]
                elif symstring == 'permute':
                    # Compute cycle type of full permutation group:
                    symmetry = permute_symmetry(nhood)
                elif symstring == 'none':
                    symmetry = []
                else:
                    symmetry = literal_eval(symstring)
                # Convert cycles to permutations:
                symmetry = [cycles2perm(numParams, x) for x in symmetry]
            else:
                read_transition(line, variables, transitions, used, numParams, symmetry)

        elif line[0:4] == 'var ':

            line = line[4:].strip() # strip var keyword
            # read each variable into a dictionary mapping string to list of ints
            line = line.replace('=',' ').replace('{',' ').replace(',',' ')
            entries = line.replace(':',' ').replace('}',' ').split()
            curvar = set([])
            for e in entries[1:]:
                if e in variables:
                    curvar.update(variables[e])
                else:
                    curvar.add(int(e))
            variables[entries[0]] = sorted(curvar)

        else:

            # assume line is a transition
            read_transition(line, variables, transitions, used, numParams, symmetry)

    return n_states, nhood, transitions



# Determine linear automorphisms of a finite set of integer points

from itertools import combinations, permutations

def subdets(pts):
    '''
    Compute determinants of all (n choose d) size-d subsets of n
    points in Z^d.
    '''

    n = len(pts)
    d = len(pts[0])
    s = {tuple([]): 1}

    for e in range(1, d+1):
        for t in combinations(range(n), e):
            t = tuple(t)

            sgn = 1
            val = 0

            for i in range(e):
                subt = t[:i] + t[i+1:]
                val += s[subt] * pts[t[i]][-e] * sgn
                sgn = 0 - sgn

            s[t] = val

    return s

def adjugate_row(pts, j):

    n = len(pts)
    s = subdets([p[:j] + p[j+1:] for p in pts])
    t = tuple(range(n))
    for i in range(n):
        subt = t[:i] + t[i+1:]
        sgn = 1 - (((i + j) % 2) * 2)
        yield s[subt] * sgn

def adjugate(mat):
    '''
    The inverse of a unimodular matrix
    '''
    return [list(adjugate_row(mat, j)) for j in range(len(mat))]

def matperms(pts):

    n = len(pts)
    d = len(pts[0])

    s = subdets(pts)
    adj = None
    adjt = None
    basis = None
    ptsset = None
    origdet = None
    minofdet = None

    for t in combinations(range(n), d):
        t = tuple(t)
        absdet = abs(s[t])
        if absdet == 0:
            continue

        if minofdet is None:
            minofdet = absdet
        else:
            minofdet = min(absdet, minofdet)

    for t in combinations(range(n), d):
        t = tuple(t)

        det = s[t]
        if abs(det) == minofdet:
            coords = [pts[i] for i in t]
            if adj is None:
                origdet = det
                basis = coords
                adj = adjugate(basis)
                adjt = [[a[i] for a in adj] for i in range(d)]
                # Transform by adjugate matrix:
                pts = [[sum([x*y for (x, y) in zip(a, m)]) for a in adjt] for m in pts]
                ptsset = {tuple([x*det for x in p]): i for (i, p) in enumerate(pts)}

            for t in permutations(t):
                t = tuple(t)
                coordst = [[pts[i][j] for i in t] for j in range(d)]

                perm = []
                for p in pts:
                    q = tuple([sum([x*y for (x, y) in zip(a, p)]) for a in coordst])
                    if q in ptsset:
                        perm.append(ptsset[q])
                    else:
                        break

                if len(perm) == len(pts):
                    yield perm




# This file defines the genera accepted by lifelib. Earlier genera take
# precedence over later genera, so (for example) the rule b38s23 belongs
# to the genus 'lifelike' rather than 'isotropic'.

genus_list = []
genus_list.append({'name': 'b3s23life', 'regex': 'b3s23'})
genus_list.append({'name': 'lifelike', 'regex': 'b0?1?2?3?4?5?6?7?8?s0?1?2?3?4?5?6?7?8?'})
genus_list.append({'name': 'generations', 'regex': 'g[1-9][0-9]*b1?2?3?4?5?6?7?8?s0?1?2?3?4?5?6?7?8?'})
genus_list.append({'name': 'isotropic', 'regex': 'b[1-8ceaiknjqrytwz-]*s[0-8ceaiknjqrytwz-]*'})
genus_list.append({'name': 'isotropic-hex-ot', 'regex': 'b[1-6]*s[0-6]*h'})
genus_list.append({'name': 'isotropic-hex', 'regex': 'b[1-6omp-]*s[0-6omp-]*h'})
genus_list.append({'name': 'isotropic_other', 'regex': 'b[0-8ceaiknjqrytwz-]*s[0-8ceaiknjqrytwz-]*'})
genus_list.append({'name': 'isotropic_other-hex', 'regex': 'b[0-6omp-]*s[0-6omp-]*h'})
genus_list.append({'name': 'isotropic_other-fc', 'regex': 'b[0-8ceaiknjqrytwz-]*s[0-8ceaiknjqrytwz-]*fc'})
genus_list.append({'name': 'isotropic_other-fe', 'regex': 'b[0-8ceaiknjqrytwz-]*s[0-8ceaiknjqrytwz-]*fe'})
genus_list.append({'name': 'isotropic_other-r2vn', 'regex': 'b([0-9]+x-([0-8]([xaceijknqrtwyz][acdefgijklmnopqr])+)+|[0-8]([xaceijknqrtwyz][acdefgijklmnopqr])+|[0-9]+x)*s([0-9]+x-([0-8]([xaceijknqrtwyz][acdefgijklmnopqr])+)+|[0-8]([xaceijknqrtwyz][acdefgijklmnopqr])+|[0-9]+x)*v2'})
genus_list.append({'name': 'ltl', 'regex': 'r[234567]b[1-9][0-9]*t[1-9][0-9]*s[1-9][0-9]*t[1-9][0-9]*'})
genus_list.append({'name': 'gltl', 'regex': 'g[1-9][0-9]*r[234567]b[1-9][0-9]*t[1-9][0-9]*s[1-9][0-9]*t[1-9][0-9]*'})
genus_list.append({'name': 'isogeny', 'regex': 'g[1-9][0-9]*b[1-8ceaiknjqrytwz-]*s[0-8ceaiknjqrytwz-]*'})
genus_list.append({'name': 'isogeny-hex-ot', 'regex': 'g[1-9][0-9]*b[1-6]*s[0-6]*h'})
genus_list.append({'name': 'isogeny-hex', 'regex': 'g[1-9][0-9]*b[1-6omp-]*s[0-6omp-]*h'})
genus_list.append({'name': 'isogeny_other', 'regex': 'g[1-9][0-9]*b[0-8ceaiknjqrytwz-]*s[0-8ceaiknjqrytwz-]*'})
genus_list.append({'name': 'isogeny_other-hex', 'regex': 'g[1-9][0-9]*b[0-6omp-]*s[0-6omp-]*h'})
genus_list.append({'name': 'isogeny_other-fc', 'regex': 'g[1-9][0-9]*b[1-8ceaiknjqrytwz-]*s[0-8ceaiknjqrytwz-]*fc'})
genus_list.append({'name': 'isogeny_other-fe', 'regex': 'g[1-9][0-9]*b[1-8ceaiknjqrytwz-]*s[0-8ceaiknjqrytwz-]*fe'})
genus_list.append({'name': 'isogeny_other-r2vn', 'regex': 'g[1-9][0-9]*b([0-9]+x-([0-8]([xaceijknqrtwyz][acdefgijklmnopqr])+)+|[0-8]([xaceijknqrtwyz][acdefgijklmnopqr])+|[0-9]+x)*s([0-9]+x-([0-8]([xaceijknqrtwyz][acdefgijklmnopqr])+)+|[0-8]([xaceijknqrtwyz][acdefgijklmnopqr])+|[0-9]+x)*v2'})
genus_list.append({'name': 'bsfkl', 'regex': 'b1?2?3?4?5?6?7?8?s0?1?2?3?4?5?6?7?8?f0?1?2?3?4?5?6?7?8?k0?1?2?3?4?5?6?7?8?l0?1?2?3?4?5?6?7?8?'})
genus_list.append({'name': 'hrot', 'regex': 'r[2345]b[0-9a-f]*s[0-9a-f]*z?'})
genus_list.append({'name': 'ghrot', 'regex': 'g[1-9][0-9]*r[2345]b[0-9a-f]*s[0-9a-f]*z?'})
genus_list.append({'name': 'deficient', 'regex': 'b[1-8ceaiknjqrytwz-]*s[0-8ceaiknjqrytwz-]*dp?'})
genus_list.append({'name': 'genext', 'regex': 'b[1-8ceaiknjqrytwz-]*s[0-8ceaiknjqrytwz-]*d(0[-])?[1-9][0-9]*([-][1-9][0-9]*)*'})

# This one must always be last as a 'catch-all' for custom rules with
# no more than 256 states:

genus_list.append({'name': 'eightbit', 'regex': 'x[0-9a-z_-]*'})

# A single genus (such as 'isotropic') can be associated with multiple
# rule families (square isotropic and hexagonal isotropic). We use a
# hyphen to provide an additional suffix for the benefit of Catagolue.

for g in genus_list:
    g['fullname'] = g['name']
    g['name'] = g['fullname'].split('-')[0]




from ._iwriter import iwriter, rule2gates

family = 0
bitplanes = 1

def mantissa(rulestring):

    if 'b0' in rulestring:
        return {0, 2, 4, 6, 8}
    else:
        return {0, 1, 2, 3, 4, 5, 6, 7, 8}

def create_rule(rulestring):

    if ('b0' in rulestring) and (rulestring[-1] == '8'):
        raise ValueError('B0 and S8 are incompatible; please invert states.')

    logstring = rulestring[rulestring.index('b'):]
    for iset in [['sse2'], ['sse2', 'avx'], ['sse2', 'avx', 'avx2']]:
        with open('ll_%s_%s.asm' % (iset[-1], logstring), 'w') as f:
            ix = iwriter(f, iset)
            ix.genlogic(logstring)

    with open('iterators_gpu.h', 'w') as f:
        f.write('''#define ADVANCE_TILE_64(a, b, tmp, dst) {              \\
    uint64_cu al = ROTL64(a, 1);                       \\
    uint64_cu ar = ROTR64(a, 1);                       \\
    uint64_cu xor2 = al ^ ar;                          \\
    (dst)[threadIdx.x] = xor2 ^ a;                     \\
    __syncthreads();                                   \\
    uint64_cu uda = (dst)[u] & (dst)[d];               \\
    uint64_cu udx = (dst)[u] ^ (dst)[d];               \\
    uint64_cu xmm8 = ((dst)[threadIdx.x] & udx) | uda; \\
    uint64_cu xmm10 = (dst)[threadIdx.x] ^ udx;        \\
    (tmp)[threadIdx.x] = (al & ar) | (a & xor2);       \\
    __syncthreads();                                   \\
    udx = (tmp)[u] ^ (tmp)[d];                         \\
    uda = (tmp)[u] & (tmp)[d];                         \\
    uint64_cu xmm9 = ((tmp)[threadIdx.x] & udx) | uda; \\
    uint64_cu xmm11 = (tmp)[threadIdx.x] ^ udx;        \\
    xor2 = xmm8 & xmm11;                               \\
    uda = a;                                           \\
    xmm8 ^= xmm11;                                     \\\n''')

        rchars, negate, beexor, essxor = rule2gates(rulestring)
        usetopbit = (essxor or beexor)

        regnames = ["xmm10", "xmm8", "xmm9", "uda", "xor2", "udx"]
        opnames = [" & ", " | ", " & ~", "nonsense", " ^ "]

        if (usetopbit):
            f.write('''    xmm11 = xor2 & xmm9;                               \\\n''')
        if (essxor and not beexor):
            f.write('''    xmm11 &= a;                                        \\\n''')
        if (beexor and not essxor):
            f.write('''    xmm11 &= (~a);                                     \\\n''')

        f.write('''    xmm9 ^= xor2;                                      \\\n''')

        for i in range(0, len(rchars), 4):
            f.write('    %s = %s%s%s; \\\n' % (regnames[rchars[i]], regnames[rchars[i+1]], opnames[rchars[i+3]], regnames[rchars[i+2]]))

        if usetopbit:
            f.write('''    xmm10 ^= xmm11;                                    \\\n''')

        if negate:
            f.write('''    b = ~xmm10;                                        \\\n}\n\n''')
        else:
            f.write('''    b = xmm10;                                         \\\n}\n\n''')

    with open('iterators_%s.h' % rulestring, 'w') as f:
        f.write('#pragma once\n')
        f.write('#include <stdint.h>\n')
        f.write('#include "../lifeconsts.h"\n')
        f.write('#include "../lifeperm.h"\n')
        f.write('#include "../eors.h"\n')
        f.write('namespace %s {\n\n' % rulestring.replace('-', '_'))

        for iset in [['sse2'], ['sse2', 'avx'], ['sse2', 'avx', 'avx2']]:
            iw = iwriter(f, iset)
            iw.write_function(rulestring, 32, 28)
            iw.write_function(rulestring, 28, 24)
            iw.write_function(rulestring, 24, 20)
            iw.write_function(rulestring, 20, 16)
            iw.write_function(rulestring, 16, 12)
            iw.write_function(rulestring, 12, 8)
            iw.write_iterator()

        f.write('\n#include "../leaf_iterators.h"\n')
        f.write('}\n')




import os
import subprocess
from ctypes import *
from numbers import Integral

try:
    import cPickle as pickle
except ImportError:
    import pickle

import base64
import re

__all__ = ['tzcount', 'call_underlying', 'WrappedLibrary', 'binstring', 'regstring', 'restypes', 'apilines']

pythlib_dir = os.path.dirname(os.path.abspath(__file__))
lifelib_dir = os.path.abspath(os.path.join(pythlib_dir, '..'))

with open(os.path.join(lifelib_dir, 'lifelib.cpp')) as f:

    apilines = [x.replace('{', ';') for x in f if re.match('^    [a-z]', x)]
    apitypes = dict([x.split('(')[0].strip().split()[::-1] for x in apilines])
    api2res = {'void': None, 'void*': c_void_p, 'bool': c_bool, 'int64_t': c_int64, 'uint64_t': c_uint64, 'int': None}
    restypes = {k : api2res[v] for (k, v) in apitypes.items()}
    apilines = ''.join(apilines)

def binstring(x):

    if x is None:
        return binstring('')
    elif type(x).__name__ == 'bytes':
        return x
    else:
        return x.encode('utf-8')

def make_compatible(x):

    if isinstance(x, tuple) and (len(x) == 1):
        return c_void_p(x[0])
    elif isinstance(x, list) and (len(x) == 1):
        return create_string_buffer(x[0])
    elif isinstance(x, list) and (len(x) >= 2):
        return ([c_uint64, c_int64, c_uint32, c_int32][x[0]] * x[1])()
    elif hasattr(x, 'ctypes'):
        import numpy as np
        return np.ascontiguousarray(x)
    elif hasattr(x, 'encode'):
        return x.encode('utf-8')
    else:
        return x

def regstring(x):

    if isinstance(x, str):
        return x
    else:
        return x.decode('utf-8')

def tzcount(x):

    return bin(x)[::-1].find('1')

def ensure_stack(required):

    if (required % 4096):
        required += (4096 - (required % 4096))

    try:
        import resource
        sl, hl = resource.getrlimit(resource.RLIMIT_STACK)
        if 0 < sl < required: # Pythonic chained comparators
            resource.setrlimit(resource.RLIMIT_STACK, (required, hl))
            sl, hl = resource.getrlimit(resource.RLIMIT_STACK)
        return (sl, hl)
    except Exception as e:
        return None


def call_underlying(lifelib, fname_and_args):

    fname = regstring(fname_and_args[0])
    args = fname_and_args[1:]

    if (fname == 'EnsureStack'):
        return ensure_stack(args[0])

    retpos = -1
    retsize = 0

    returnables = [list]

    for (i, x) in enumerate(args):
        if isinstance(x, list):
            retpos = i
            retsize = len(x)
        elif hasattr(x, 'ctypes'):
            retpos = i
            retsize = 2

    args = [make_compatible(x) for x in args]

    try:
        func = getattr(lifelib, fname)
    except AttributeError:
        if 'Version' in fname:
            return "old"
        else:
            raise

    rt = restypes[fname]
    if rt is not None:
        func.restype = rt

    def f(i, a):
        if hasattr(a, 'ctypes'):
            return c_void_p(a.ctypes.data)
        elif (retpos == i) and (retsize == 1):
            return cast(a, c_void_p)
        else:
            return a

    retval = func(*[f(i, a) for (i, a) in enumerate(args)])

    if (rt is not None) and (rt == c_void_p):
        if isinstance(retval, c_void_p):
            retval = tuple([retval.value])
        else:
            retval = tuple([retval])
    elif (retpos >= 0):
        if hasattr(args[retpos], 'ctypes'):
            retval = args[retpos]
        elif (retsize == 1):
            retval = regstring(args[retpos].value)
        else:
            retval = list(args[retpos])

    return retval


def remote_load():
    '''
    Minimalistic function which operates lifelib from instructions received
    from stdin. Results and exceptions are returned via stdout.
    '''

    from sys import argv, stdin, stdout, stderr
    from traceback import format_exc

    if hasattr(stdin, 'buffer'):
        stdin = stdin.buffer

    if hasattr(stdout, 'buffer'):
        stdout = stdout.buffer

    the_library = cdll.LoadLibrary(argv[1])

    while True:

        obj = pickle.load(stdin)

        if (len(obj[0]) <= 1):
            break

        try:
            retval = (0, call_underlying(the_library, obj))
        except Exception:
            retval = (1, format_exc())

        pickle.dump(retval, stdout, 2)
        stdout.flush()

    pickle.dump((0, 0), stdout, 2)
    stdout.flush()


class WrappedLibrary(object):

    def __init__(self, soname, local_bash=None, local_python='python'):

        self.last_power = 800
        soname = os.path.abspath(soname)

        if local_bash is None:
            self.the_library = cdll.LoadLibrary(soname)
            self.remote = False
        else:
            args = local_bash + [local_python, 'pythlib/lowlevel.py', soname]
            self.the_library = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
            self.remote = True

    @property
    def version(self):

        v = self('GetCompiledVersion', [2048]).lower()
        while (len(v) >= 1) and (v[0] == 'l'):
            v = v[1:]

        return v

    def __call__(self, *args):

        if self.remote:
            pickle.dump(tuple(args), self.the_library.stdin, 2)
            self.the_library.stdin.flush()
            retval = pickle.load(self.the_library.stdout)

            if retval[0] == 0:
                return retval[1]
            else:
                raise RuntimeError("Child process raised exception: %s" % str(retval[1]))
        else:
            return call_underlying(self.the_library, args)

    def ensure_limit(self, exponent, mantissa, pname='Parameter'):

        powerof2 = exponent + len(bin(abs(mantissa) - 1)) - 2

        if (powerof2 > 65536):
            raise ValueError("%s exceeds maximum of 2 ^^ 5" % pname)

        if (powerof2 > self.last_power):
            x = self('EnsureStack', powerof2 * 768)
            if x is None:
                import warnings
                warnings.warn("Could not ensure stack size, probably due to running on Windows. Expect segfaults soon.")
            self.last_power = powerof2

    def annihilate(self):

        if self.remote:
            try:
                self.__call__('X', 0)
            except Exception:
                pass
            self.the_library.wait()
            self.the_library.stdin.close()
            self.the_library.stdout.close()
            self.remote = False

    def __del__(self):

        self.annihilate()

if __name__ == '__main__':

    remote_load()




import unittest
import lifelib

from lifelib.tests.integration import test_gol as tg

class TestIndirection(tg.TestGoL):

    @classmethod
    def setUpClass(cls):

        lifelib.load_rules('b3s23', force_indirect=True)

if __name__ == '__main__':
    unittest.main()




class iwriter_base(object):

    def __init__(self, f, iset):
        self.f = f
        self.iset = iset
        for k in ['sse2', 'sse3', 'ssse3', 'sse4', 'avx', 'avx2', 'avx512']:
            if k in iset:
                self.besti = k

    def printinstr(self, s):

        self.f.write('        "%s \\n\\t"\n' % s)

    def printcomment(self, s):

        self.f.write('        // %s \n' % s)

    def preparethings(self, dwidth, reg13=13, reg14=14):

        if ('avx512' in self.iset):
            self.printinstr('vmovdqu64 (%2), %%zmm14')
            self.printinstr('vmovdqu64 64(%2), %%zmm16')
            self.printinstr('vmovdqu64 128(%2), %%zmm17')
        elif ('avx2' in self.iset):
            self.printinstr('vmovdqu (%2), %%ymm'+str(reg14))
            self.printinstr('vmovdqu 192(%2), %%ymm'+str(reg13))
        else:
            r13 = '%%xmm' + str(reg13)
            r14 = '%%xmm' + str(reg14)
            shuf = 'vpshufd' if ('avx' in self.iset) else 'pshufd'
            self.printinstr('mov $0xffffffff, %%ebx')
            self.printinstr('movd %%ebx, ' + r13)
            if (dwidth):
                ddic = {28: '0x3ffffffc',
                        24: '0x0ffffff0',
                        20: '0x03ffffc0',
                        16: '0x00ffff00',
                        12: '0x003ffc00',
                         8: '0x000ff000',
                         4: '0x0003c000'}
                self.printinstr('mov $' + ddic[dwidth] + ', %%ebx')
                self.printinstr('movd %%ebx, ' + r14)
            self.printinstr('%s $1, %s, %s' % (shuf, r13, r13))
            self.printinstr('%s $0, %s, %s' % (shuf, r14, r14))

    def trogicgate(self, op, inreg1, inreg2, inreg3, outreg=None):

        if outreg is None:
            outreg = inreg3

        if (outreg != inreg3):
            self.printinstr('vmovdqa32 %s, %s' % ('%%zmm' + str(inreg3), '%%zmm' + str(outreg)))

        self.printinstr('vpternlogd $%d, %s, %s, %s' % (op, '%%zmm' + str(inreg1), '%%zmm' + str(inreg2), '%%zmm' + str(outreg)))

    def logicgate(self, op, inreg1, inreg2, outreg, regname=None):

        if regname is None:
            regname = '%%zmm' if ('avx512' in self.iset) else ('%%ymm' if ('avx2' in self.iset) else '%%xmm')
        i1 = regname + str(inreg1)
        i2 = regname + str(inreg2)
        o1 = regname + str(outreg)

        if 'avx' in self.iset:
            self.printinstr('v%s %s, %s, %s' % (op, i1, i2, o1))
        elif (i2 == o1):
            self.printinstr('%s %s, %s' % (op, i1, o1))
        elif (i1 == o1):
            if (op == 'pandn'):
                self.printinstr('por %s, %s' % (i2, o1))
                self.printinstr('pxor %s, %s' % (i2, o1))
            else:
                self.printinstr('%s %s, %s' % (op, i2, o1))
        else:
            self.printinstr('movdqa %s, %s' % (i2, o1))
            self.printinstr('%s %s, %s' % (op, i1, o1))

    def write16n(self, n, inreg, offset, outloc):

        inregz = '%%zmm' + str(inreg)
        inregy = '%%ymm' + str(inreg)
        inregx = '%%xmm' + str(inreg)

        reg16z = '%%zmm13'
        reg16y = '%%ymm13'
        reg16x = '%%xmm13'

        d1 = ('' if (offset == 0) else str(offset)) + outloc
        d2 = str(offset + 32) + outloc

        if (n == 16):
            self.printinstr('vmovdqu %s, %s' % (inregx, d1))
        elif (n == 32):
            self.printinstr('vmovdqu %s, %s' % (inregy, d1))
        elif (n == 48):
            self.printinstr('vshufi32x4 $78, %s, %s, %s' % (inregz, inregz, reg16z))
            self.printinstr('vmovdqu %s, %s' % (inregy, d1))
            self.printinstr('vmovdqu %s, %s' % (reg16x, d2))
        elif (n == 64):
            self.printinstr('vmovdqu64 %s, %s' % (inregz, d1))

    def read16n(self, n, inloc, offset, outreg):

        outregz = '%%zmm' + str(outreg)
        outregy = '%%ymm' + str(outreg)
        outregx = '%%xmm' + str(outreg)

        reg16z = '%%zmm13'
        reg16y = '%%ymm13'
        reg16x = '%%xmm13'

        d1 = ('' if (offset == 0) else str(offset)) + inloc
        d2 = str(offset + 32) + inloc

        if (n == 16):
            self.printinstr('vmovdqu %s, %s' % (d1, outregx))
        elif (n == 32):
            self.printinstr('vmovdqu %s, %s' % (d1, outregy))
        elif (n == 48):
            self.printinstr('vmovdqu %s, %s' % (d1, outregy))
            self.printinstr('vmovdqu %s, %s' % (d2, reg16x))
            self.printinstr('vshufi32x4 $68, %s, %s, %s' % (reg16z, outregz, outregz))
        elif (n == 64):
            self.printinstr('vmovdqu64 %s, %s' % (d1, outregz))

    def write_function(self, rulestring, rowcount, dwidth):

        name = 'iterate_%s_%d_%d' % (self.besti, rowcount, dwidth)

        params = 'uint32_t * __restrict__ diffs, bool onegen'
        for i in 'jhed':
            params = 'uint32_t * __restrict__ ' + i + ', ' + params

        self.f.write('    bool %s(%s) {\n' % (name, params))

        self.f.write('        if (h) {\n')
        self.f.write('            for (int i = 0; i < %d; i++) {\n' % (rowcount))
        self.f.write('                h[i] |= d[i];\n')
        self.f.write('            }\n')
        self.f.write('        }\n')

        self.f.write('        if (j) {\n')
        self.f.write('            for (int i = 0; i < %d; i++) {\n' % (rowcount))
        self.f.write('                j[i] &= d[i];\n')
        self.f.write('            }\n')
        self.f.write('        }\n')

        self.assemble(rulestring, 0, rowcount, dwidth)

        self.f.write('        if (h) {\n')
        self.f.write('            for (int i = 1; i < %d; i++) {\n' % (rowcount - 1))
        if (rulestring[:2] == 'b0'):
            # We want the history state to match the envelope of the Gollyfied
            # version of the B0 rule:
            self.f.write('                h[i] |= (~e[i-1]);\n')
        else:
            self.f.write('                h[i] |= e[i-1];\n')
        self.f.write('            }\n')
        self.f.write('        }\n')

        self.f.write('        if (j) {\n')
        self.f.write('            for (int i = 1; i < %d; i++) {\n' % (rowcount - 1))
        self.f.write('                j[i] &= e[i-1];\n')
        self.f.write('            }\n')
        self.f.write('        }\n')

        self.f.write('        if (onegen) {\n')
        self.f.write('            for (int i = 2; i < %d; i++) {\n' % (rowcount - 2))
        self.f.write('                d[i] = e[i-1];\n')
        self.f.write('            }\n')
        self.f.write('            return false;\n')
        self.f.write('        }\n')

        self.assemble(rulestring, 1, rowcount, dwidth)

        self.f.write('        if (h) {\n')
        self.f.write('            for (int i = 2; i < %d; i++) {\n' % (rowcount - 2))
        self.f.write('                h[i] |= d[i];\n')
        self.f.write('            }\n')
        self.f.write('        }\n')

        self.f.write('        if (j) {\n')
        self.f.write('            for (int i = 2; i < %d; i++) {\n' % (rowcount - 2))
        self.f.write('                j[i] &= d[i];\n')
        self.f.write('            }\n')
        self.f.write('        }\n')

        if 'avx512' in self.iset:
            newrows = rowcount - 4

            while (newrows >= 32):
                newrows -= 16

            if (newrows <= 16):
                bindices = list(range(0, newrows // 2))
            elif (newrows == 28):
                bindices = [8, 9, 10, 11, 12, 13]
            else:
                bindices = list(range(4, newrows // 2))
            lindices = list(range(newrows - 2, newrows))

        elif 'avx2' in self.iset:

            bindices = [4, 5, 6, 7]
            lindices = list(range(16 + ((rowcount - 6) % 8), 18 + ((rowcount - 6) % 8)))

        else:

            bindices = [2, 3]
            lindices = list(range(8 + ((rowcount - 6) % 4), 10 + ((rowcount - 6) % 4)))

        self.f.write('        uint64_t* e64 = ((uint64_t*) e);\n')
        self.f.write('        uint64_t bigdiff = %s;\n' % (' | '.join(['e64[%d]' % x for x in bindices])))
        self.f.write('        if (diffs != 0) {\n')
        self.f.write('            diffs[0] = (bigdiff | (bigdiff >> 32));\n')
        self.f.write('            diffs[1] = e[0] | e[1];\n')
        self.f.write('            diffs[2] = %s;\n' % (' | '.join(['e[%d]' % x for x in lindices])))
        self.f.write('        }\n')
        self.f.write('        return (bigdiff == 0);\n')
        self.f.write('    }\n\n')

    def write_iterator(self):

        name = 'iterate_var_%s' % self.besti
        self.f.write('    int %s(int n, uint32_t * __restrict__ d, uint32_t * __restrict__ h, uint32_t * __restrict__ j) {\n' % name)
        self.f.write('        uint32_t e[32];\n')
        self.f.write('        if (n >= 7) { if (iterate_%s_32_28(d, e, h, j, 0, (n == 7))) {return 8;} }\n' % self.besti)
        self.f.write('        if (n >= 5) { if (iterate_%s_28_24(d+2, e+2, h+2, j+2, 0, (n == 5))) {return 6;} }\n' % self.besti)
        self.f.write('        if (n >= 3) { if (iterate_%s_24_20(d+4, e+4, h+4, j+4, 0, (n == 3))) {return 4;} }\n' % self.besti)
        self.f.write('        if (n >= 1) { if (iterate_%s_20_16(d+6, e+6, h+6, j+6, 0, (n == 1))) {return 2;} }\n' % self.besti)
        self.f.write('        return 0;\n')
        self.f.write('    }\n\n')

        self.f.write('    int %s(int n, uint32_t * __restrict__ d, uint32_t * __restrict__ h) {\n' % name)
        self.f.write('        uint32_t e[32];\n')
        self.f.write('        if (n >= 7) { if (iterate_%s_32_28(d, e, h, 0, 0, (n == 7))) {return 8;} }\n' % self.besti)
        self.f.write('        if (n >= 5) { if (iterate_%s_28_24(d+2, e+2, h+2, 0, 0, (n == 5))) {return 6;} }\n' % self.besti)
        self.f.write('        if (n >= 3) { if (iterate_%s_24_20(d+4, e+4, h+4, 0, 0, (n == 3))) {return 4;} }\n' % self.besti)
        self.f.write('        if (n >= 1) { if (iterate_%s_20_16(d+6, e+6, h+6, 0, 0, (n == 1))) {return 2;} }\n' % self.besti)
        self.f.write('        return 0;\n')
        self.f.write('    }\n\n')

        self.f.write('    int %s(int n, uint32_t * __restrict__ d) {\n' % name)
        self.f.write('        uint32_t e[32];\n')
        self.f.write('        if (n >= 7) { if (iterate_%s_32_28(d, e, 0, 0, 0, (n == 7))) {return 8;} }\n' % self.besti)
        self.f.write('        if (n >= 5) { if (iterate_%s_28_24(d+2, e+2, 0, 0, 0, (n == 5))) {return 6;} }\n' % self.besti)
        self.f.write('        if (n >= 3) { if (iterate_%s_24_20(d+4, e+4, 0, 0, 0, (n == 3))) {return 4;} }\n' % self.besti)
        self.f.write('        if (n >= 1) { if (iterate_%s_20_16(d+6, e+6, 0, 0, 0, (n == 1))) {return 2;} }\n' % self.besti)
        self.f.write('        return 0;\n')
        self.f.write('    }\n\n')


def rule2gates(rulestring):

    bee = [0] * 10
    ess = [0] * 10

    for char in rulestring:
        if (char == 'b'):
            birth = True
        elif (char == 's'):
            birth = False
        else:
            k = int(char)
            if (birth):
                bee[k] = 1
            else:
                ess[k+1] = 1

    negate = bee[0]
    beexor = (bee[0] != bee[8])
    essxor = (ess[1] != ess[9])

    stars = ("*" if essxor else "") + ("**" if beexor else "") + ("!" if negate else "")

    if negate:
        bee = [1 - x for x in bee]
        ess = [1 - x for x in ess]

    ess[0] = (1 - ess[8]) if essxor else ess[8]

    ruleint = 0;
    for i in range(8):
        ruleint += (bee[i] << i)
        ruleint += (ess[i] << (i + 8))

    rident = None

    with open('../boolean.out', 'r') as f:
        for fline in f:
            x = fline.split(":  ")
            if (len(x) == 2) and (int(x[0]) == ruleint):
                rident = x[1][:-1]

    if (ruleint == 0):
        rident = '-004'
    if (ruleint == 65280):
        rident = '-331'
    if (ruleint == 61680):
        rident = '-221'
    if (ruleint == 52428):
        rident = '-111'
    if (ruleint == 43690):
        rident = '-001'

    if rident is None:
        raise ValueError("Error: unrecognised rule")
    # print(("Rule circuit:     ["+rident+"]"+stars))

    rchars = list(rident)
    for i in range(len(rchars)):
        if (rchars[i] == '-'):
            rchars[i] = (i//4) + 4
        else:
            rchars[i] = int(rchars[i])

    for i in range(0, len(rchars), 4):
        if (rchars[i+3] == 3):
            rchars[i+3] = 2
            rchars[i+2] ^= rchars[i+1]
            rchars[i+1] ^= rchars[i+2]
            rchars[i+2] ^= rchars[i+1]

    # Map logical registers to physical registers:
    for i in range(0, len(rchars), 4):
        dependencies = [False] * 10
        for j in range(i+4, len(rchars), 4):
            dependencies[rchars[j+1]] = True
            dependencies[rchars[j+2]] = True
        d = -1
        e = rchars[i]
        for j in range(6):
            if (e == j):
                d = j
                break
            if not dependencies[j]:
                d = j
                break
        if (d == -1):
            raise RuntimeError("Error: insufficiently many physical registers")
        else:
            # print(str(e)+" --> "+str(d))
            rchars[i] = d
            for j in range(i+4, len(rchars), 4):
                if (rchars[j+1] == e):
                    rchars[j+1] = d
                if (rchars[j+2] == e):
                    rchars[j+2] = d

    return (rchars, negate, beexor, essxor)


class iwriter(iwriter_base):

    def load_and_hshift(self, i, oddgen, terminal):

        regbytes = 32 if ('avx2' in self.iset) else 16
        regname = '%%ymm' if (('avx2' in self.iset) and not terminal) else '%%xmm'
        accessor = 'vmovdqu' if ('avx' in self.iset) else 'movups'
        inreg = regname + str(5 - 3 * (i % 2))

        d = '(%1)' if (oddgen) else '(%0)'
        d = d if (i == 0) else (str(regbytes * i) + d)

        self.printinstr('%s %s, %s' % (accessor, d, inreg))
        if ('avx' in self.iset):
            self.printinstr('vpsrld $1, %s, %s0' % (inreg, regname))
            self.printinstr('vpslld $1, %s, %s1' % (inreg, regname))
        else:
            self.printinstr('movdqa %s, %s0' % (inreg, regname))
            self.printinstr('movdqa %s, %s1' % (inreg, regname))
            self.printinstr('psrld $1, %s0' % regname)
            self.printinstr('pslld $1, %s1' % regname)

    def horizontal_adders(self, i):

        self.logicgate('pxor', 0, 1, 6 - 3 * (i % 2))
        self.logicgate('pand', 0, 1, 7 - 3 * (i % 2))
        self.logicgate('pand', 5 - 3 * (i % 2), 6 - 3 * (i % 2), 1)
        self.logicgate('pxor', 5 - 3 * (i % 2), 6 - 3 * (i % 2), 6 - 3 * (i % 2))
        self.logicgate('por', 1, 7 - 3 * (i % 2), 7 - 3 * (i % 2))

    def vertical_bitshifts(self, i):

        if 'avx2' in self.iset:
            self.logicgate('pblendd $1,', 6 - 3 * (i % 2), 3 + 3 * (i % 2), 8)
            self.logicgate('pblendd $1,', 7 - 3 * (i % 2), 4 + 3 * (i % 2), 9)
            self.logicgate('pblendd $3,', 6 - 3 * (i % 2), 3 + 3 * (i % 2), 10)
            self.logicgate('pblendd $3,', 7 - 3 * (i % 2), 4 + 3 * (i % 2), 11)
            self.logicgate('pblendd $1,', 5 - 3 * (i % 2), 2 + 3 * (i % 2), 12)
            self.logicgate('permd', 8, 13, 8)
            self.logicgate('permd', 9, 13, 9)
            self.printinstr('vpermq $57, %%ymm10, %%ymm10')
            self.printinstr('vpermq $57, %%ymm11, %%ymm11')
            self.logicgate('permd', 12, 13, 12)
        else:
            self.logicgate('pand', 13, 3 + 3 * (i % 2), 8)
            self.logicgate('pand', 13, 4 + 3 * (i % 2), 9)
            self.logicgate('pand', 13, 2 + 3 * (i % 2), 12)
            self.logicgate('pandn', 6 - 3 * (i % 2), 13, 0)
            self.logicgate('por', 0, 8, 8)
            self.logicgate('pandn', 7 - 3 * (i % 2), 13, 0)
            self.logicgate('por', 0, 9, 9)
            self.logicgate('pandn', 5 - 3 * (i % 2), 13, 0)
            self.logicgate('por', 0, 12, 12)
            self.logicgate('shufps $0x39,', 8, 8, 8)
            self.logicgate('shufps $0x39,', 9, 9, 9)
            self.logicgate('shufps $0x4e,', 6 - 3 * (i % 2), 3 + 3 * (i % 2), 10)
            self.logicgate('shufps $0x4e,', 7 - 3 * (i % 2), 4 + 3 * (i % 2), 11)
            self.logicgate('shufps $0x39,', 12, 12, 12)

    def vertical_adders(self, i):

        self.logicgate('pxor', 3 + 3 * (i % 2), 8, 8)
        self.logicgate('pxor', 4 + 3 * (i % 2), 9, 9)
        self.logicgate('pxor', 8, 10, 10)
        self.logicgate('pxor', 9, 11, 11)
        self.logicgate('por', 8, 3 + 3 * (i % 2), 3 + 3 * (i % 2))
        self.logicgate('por', 9, 4 + 3 * (i % 2), 4 + 3 * (i % 2))
        self.logicgate('pand', 10, 8, 8)
        self.logicgate('pand', 11, 9, 9)
        self.logicgate('pandn', 3 + 3 * (i % 2), 8, 8)
        self.logicgate('pandn', 4 + 3 * (i % 2), 9, 9)

    def genlogic(self, rulestring):

        rchars, negate, beexor, essxor = rule2gates(rulestring)

        usetopbit = (essxor or beexor)

        self.logicgate('pand', 8, 11, 1)
        self.logicgate('pxor', 11, 8, 8)
        if (beexor and not essxor):
            self.logicgate('pand', 1, 9, 0)
            self.logicgate('pandn', 0, 12, 11)
        elif (usetopbit):
            self.logicgate('pand', 1, 9, 11)
        if (essxor and not beexor):
            self.logicgate('pand', 12, 11, 11)
        self.logicgate('pxor', 1, 9, 9)

        regnames = [10, 8, 9, 12, 1, 0]
        opnames = ["and", "or", "andn", "nonsense", "xor"]

        for i in range(0, len(rchars), 4):
            self.logicgate('p'+opnames[rchars[i+3]], regnames[rchars[i+1]], regnames[rchars[i+2]], regnames[rchars[i]])

        if (usetopbit):
            # printcomment(g, 'correct for B8/S8 nonsense:')
            self.logicgate('pxor', 11, 10, 10)

        if (negate):
            # Rule contains B0:
            self.logicgate('pcmpeqb', 11, 11, 11)
            self.logicgate('pxor', 11, 10, 10)


    def save_result(self, i, oddgen, terminal, diff=False):

        regbytes = 32 if ('avx2' in self.iset) else 16
        if oddgen:
            e = str(regbytes * (i - 1) + 8) + '(%0)'
        else:
            e = '(%1)' if (i == 1) else (str(regbytes * (i - 1)) + '(%1)')

        regname = '%%ymm' if (('avx2' in self.iset) and not terminal) else '%%xmm'
        accessor = 'vmovdqu' if ('avx' in self.iset) else 'movups'
        if diff:
            self.logicgate('pand', 14, 10, 10, regname)
            self.printinstr('%s %s, %s8' % (accessor, e, regname))
            self.logicgate('pandn', 8, 14, 11, regname)
            self.logicgate('por', 10, 11, 11, regname)
            self.printinstr('%s %s11, %s' % (accessor, regname, e))
            regname = '%%ymm' if ('avx2' in self.iset) else '%%xmm'
            if (diff == 'initial'):
                self.logicgate('pxor', 11, 8, 15)
                self.printinstr('%s %s15, %s' % (accessor, regname, '(%1)'))
            else:
                self.logicgate('pxor', 11, 8, 8)
                self.logicgate('por', 8, 15, 15)
            if (diff == 'final'):
                pos2 = '64(%1)' if ('avx2' in self.iset) else '32(%1)'
                pos1 = '32(%1)' if ('avx2' in self.iset) else '16(%1)'
                self.printinstr('%s %s8, %s' % (accessor, regname, pos2))
                self.printinstr('%s %s15, %s' % (accessor, regname, pos1))
        else:
            self.printinstr('%s %s10, %s' % (accessor, regname, e))

    def prologue(self):

        self.f.write('        asm (\n')

    def epilogue(self, dwidth):

        self.f.write('                : /* no output operands */ \n')
        self.f.write('                : "r" (d), "r" (e)')
        if (dwidth):
            self.f.write(', "r" (apg::__sixteen%d)' % dwidth)
        self.f.write('\n')
        self.f.write('                : "ebx", ')
        for i in range(16):
            self.f.write('"xmm%d", ' % i)
            if (i % 6 == 4):
                self.f.write('\n' + (' ' * 20))
        self.f.write('"memory");\n\n')

    def assemble(self, rulestring, oddgen, rowcount, dwidth):

        self.prologue()
        self.preparethings(dwidth)

        if ('avx2' in self.iset):
            iters = rowcount // 8 + (2 if ((rowcount % 8) and not oddgen) else 1)
            rpr = 8
        else:
            iters = rowcount // 4 + (0 if oddgen else 1)
            rpr = 4

        for i in range(iters):
            if (i * rpr < rowcount):
                terminal = ((i + 1) * rpr > rowcount)
                self.load_and_hshift(i, oddgen, terminal)
                self.horizontal_adders(i)
            if (i > 0):
                self.vertical_bitshifts(i)
                self.vertical_adders(i)
                self.f.write('#include "ll_%s_%s.asm"\n' % (self.besti, rulestring))
                terminal = (i * rpr == rowcount + (0 if oddgen else 4))
                if oddgen:
                    if (i == 1):
                        diff = 'initial'
                    elif (i + 1 == iters):
                        diff = 'final'
                    else:
                        diff = True
                else:
                    diff = False
                self.save_result(i, oddgen, terminal, diff)

        self.epilogue(dwidth)




import unittest
import lifelib
import os

class TestRuleTrees(unittest.TestCase):

    def test_wright(self):

        sess = lifelib.load_rules('b3456s235678d1-7', 'b2s34678d', force_compile=True)
        lt = sess.lifetree()
        x = lt.pattern('6.F$3.A.GFG$.2ABA2H$A2BFA.G$ABDFH$2ADBHF$2.CAGD.H$2.2ADEH$3.2ADAHC$5.2BDB$5.3A2.H$10.H!', 'b3456s235678d1-7')
        self.assertEqual(x.apgcode, 'xq24_0sc315ek8gzx1x2l1a_y0aagogzw11w2368e4_wgcco8zy223iu_xggz44x2')
        y = lt.pattern('xq83_0h_4h0a', 'b2s34678d')
        self.assertEqual(y.apgcode, 'xq83_0h_4h0a')

    def test_langtons_loops(self):

        filename = os.path.join(lifelib.lifelib_dir, 'rules', 'source', 'Langtons-Loops.table')
        sess = lifelib.load_rules(filename, force_compile=True)
        lt = sess.lifetree()
        ll = lt.pattern('''x = 15, y = 10, rule = Langtons-Loops
.8B$BAG.AD.ADB$B.6B.B$BGB4.BAB$BAB4.BAB$B.B4.BAB$BGB4.BAB$BA6BA5B$B.G
A.GA.G5AB$.13B!''')

        self.assertEqual(ll.getrule(), 'x9xlangtons-loops')

        self.assertEqual(ll.population, 86)
        self.assertEqual(ll[1000].population, 4154)
        self.assertEqual(ll[1000000].population, 7013241874)

    def test_brew(self):

        filename = os.path.join(lifelib.lifelib_dir, 'rules', 'source', 'Brew.rule')
        sess = lifelib.load_rules(filename, force_compile=True)
        lt = sess.lifetree()
        p150 = lt.pattern('''x = 30, y = 21, rule = Brew
11.2B4.2B$10.B2.B2.B2.B$11.2B4.2B2$13.4C$2.C9.C4.C9.C$3.2C4.BCB6.BCB
4.2C$2C.C5.BCBC4.CBCB5.C.2C$2.2C3.2B.2A6.2A.2B3.2C$.C4.C.AB2CBA2.AB2C
BA.C4.C$.C3.2BCA4BA2.A4BAC2B3.C$.C4.C.AB2CBA2.AB2CBA.C4.C$2.2C3.2B.2A
6.2A.2B3.2C$2C.C5.BCBC4.CBCB5.C.2C$3.2C4.BCB6.BCB4.2C$2.C9.C4.C9.C$
13.4C2$11.2B4.2B$10.B2.B2.B2.B$11.2B4.2B!''')

        self.assertEqual(p150.period, 150)
        self.assertEqual(p150.apgcode, 'xp150_y3coa4aoczy4vxvzy33152513_o4oy0cen9necy0o4ozg9gy2vfvy2g9gz121y037e9e73y0121')




def identify_compression(filename):
    '''
    Determine the file compression algorithm by inspecting the first few bytes
    of the file. This supports bz2, gzip, and lzma compression.
    '''

    with open(filename, 'rb') as f:
        x = list(f.read(8))

    if isinstance(x[0], str):
        x = [ord(a) for a in x]

    if x[:3] == [0x42, 0x5a, 0x68]:
        return 'bz2'
    elif x[:2] == [0x1f, 0x8b]:
        return 'gzip'
    elif x[:7] == [0xfd, 0x37, 0x7a, 0x58, 0x5a, 0x00, 0x00]:
        return 'lzma'
    else:
        return None

def identify_metadata(filename):
    '''
    Identifies the file format and other details about a pattern file.
    '''

    metadata = {'format': 'rle', 'comments': []}

    with open(filename, 'r') as f:
        for l in f:
            l = l.strip()
            if l == '':
                continue
            if l[0] == '#':
                if (l[:8] == '#FRAMES ') and (metadata['format'] == 'macrocell'):
                    metadata['frames'] = l[8:].split()
                else:
                    metadata['comments'].append(l)
            elif l[0] == '[':
                metadata['format'] = 'macrocell'
            else:
                break

    return metadata




import os
import lifelib
import unittest
import lifelib.genera.rulefiles.parsetree as parsetree
import lifelib.genera.rulefiles.automorph as automorph

from lifelib.genera.rulefiles import table_to_tree

class TestTreeduce(unittest.TestCase):

    def test_adjugate(self):
        '''
        Test Wikipedia's example of an adjugate matrix
        '''

        mat = [[-3, 2, -5], [-1, 0, -2], [3, -4, 1]]
        adj = [[-8, 18, -4], [-5, 12, -1], [4, -6, 2]]
        adj2 = automorph.adjugate(mat)

        self.assertEqual(adj2, adj)

    def test_matperms(self):
        '''
        Test ability to determine symmetry group of a cross.
        '''
        mp = list(automorph.matperms([[0, 0], [1, 0], [0, 1], [-1, 0], [0, -1]]))
        mp2 = [[0, 1, 2, 3, 4], [0, 2, 1, 4, 3], [0, 1, 4, 3, 2], [0, 4, 1, 2, 3],
                [0, 2, 3, 4, 1], [0, 3, 2, 1, 4], [0, 3, 4, 1, 2], [0, 4, 3, 2, 1]]

        self.assertEqual(set(map(tuple, mp)), set(map(tuple, mp2)))

        mp3 = list(automorph.matperms([[0, 0], [4, 3], [-3, 4], [-4, -3], [3, -4]]))

        self.assertEqual(set(map(tuple, mp3)), set(map(tuple, mp2)))

        mp = list(automorph.matperms([[0], [5], [-5], [7], [-7]]))

        self.assertEqual(len(mp), 2)

    def test_fcc(self):

        d3 = [(x, y, z) for x in [-1, 0, 1] for y in [-1, 0, 1] for z in [-1, 0, 1] if x*x + y*y + z*z == 2]

        mp = list(automorph.matperms(d3))

        self.assertEqual(len(mp), 48)

    def test_24cell(self):
        '''
        Tests that the vertex set of the 24-cell has a symmetry group of
        order 1152 as expected.
        '''

        t = [-1, 0, 1]
        d4 = [(w, x, y, z) for x in t for y in t for z in t for w in t if w*w + x*x + y*y + z*z == 2]

        self.assertEqual(len(set(d4)), 24)
        mp = list(automorph.matperms(d4))
        self.assertEqual(len(mp), 1152)

    def test_rotate4(self):
        '''
        Tests that we can learn the symmetries of Langton's Loops from its
        rule tree alone.
        '''

        # Load a rule table:
        filename = os.path.join(lifelib.lifelib_dir, 'rules', 'source', 'Langtons-Loops.table')
        with open(filename, 'r') as f:
            list_of_lines = list(f)

        # Convert it to a ruletree and optimise:
        list_of_lines = table_to_tree(list_of_lines)
        ruletree = parsetree.ParseRuleTree(list_of_lines)
        ruletree = parsetree.optimise_tree(*ruletree)

        # Determine that the rule has rotate4 symmetries but not reflections:
        syms = parsetree.get_symmetries(*ruletree)
        syms2 = [[0, 1, 2, 3, 4], [3, 0, 1, 2, 4], [1, 2, 3, 0, 4], [2, 3, 0, 1, 4]]
        self.assertEqual(set(map(tuple, syms)), set(map(tuple, syms2)))

    def test_reduce(self):
        '''
        Tests that we can recognise an hexagonal rule embedded as a Moore CA.
        '''

        # Load a ruletree as a list of lines:
        filename = os.path.join(lifelib.lifelib_dir, 'rules', 'source', 'Hex-B2omS2.tree')
        with open(filename, 'r') as f:
            list_of_lines = list(f)

        # Parse the ruletree and verify that it has the Moore neighbourhood:
        ruletree = parsetree.ParseRuleTree(list_of_lines)
        self.assertEqual(ruletree[1], [(-1, -1), (1, -1), (-1, 1), (1, 1), (0, -1), (-1, 0), (1, 0), (0, 1), (0, 0), (0, 0)])
        self.assertEqual(len(ruletree[2]), 31)

        # Optimise the tree and verify that the neighbourhood has reduced:
        ruletree2 = parsetree.optimise_tree(*ruletree)
        self.assertEqual(ruletree2[1], [(-1, -1), (1, 1), (0, -1), (-1, 0), (1, 0), (0, 1), (0, 0), (0, 0)])
        self.assertEqual(len(ruletree2[2]), 27)

        # Determine that the resulting rule has exactly 12 symmetries:
        syms = parsetree.get_symmetries(*ruletree2)
        self.assertEqual(len(syms), 12)

if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages
import lifelib

lifelib.reset_tree()

import os
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'lifelib', 'README.md'), 'r') as f:
    long_description = f.read()

setup(
    name='python-lifelib',
    version=lifelib.__version__,
    description='Algorithms for manipulating and simulating patterns in cellular automata',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Adam P. Goucher',
    author_email='adam@hatsya.com',
    url='https://gitlab.com/apgoucher/lifelib',
    license='MIT',
    packages=['lifelib'],
    test_suite='lifelib.tests',
    include_package_data=True,
    zip_safe=False,
    install_requires=['numpy>=1.13', 'tqdm>=4.27'],
    extras_require={'notebook': ['jupyter', 'imageio']},
    classifiers = [
        'Development Status :: 4 - Beta',

        # MIT licence as always:
        'License :: OSI Approved :: MIT License',

        # Since we require gcc or clang to compile:
        'Operating System :: POSIX',
        'Programming Language :: C++',
        'Programming Language :: Assembly',

        # Lifelib is compatible with Python 2 and 3, and has been tested on:
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',

        # Cellular automata fall into these categories:
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Life',
        'Topic :: Scientific/Engineering :: Mathematics',
    ])



'''
Isotropic generations
'''

from .generations import family, mantissa, postprocess, number_of_states, bitplanes
from .isotropic import isotrope

def create_rule(rulestring):

    nstates = number_of_states(rulestring)

    if (nstates < 3):
        raise ValueError("Number of states must be at least 3")

    logstring = rulestring[rulestring.index('b'):]
    isotrope(logstring)

    with open('iterators_%s.h' % rulestring, 'w') as f:
        f.write('#pragma once\n')
        f.write('#include <stdint.h>\n')
        f.write('#include <cstdlib>\n')
        f.write('#include <iostream>\n')
        f.write('#include "../eors.h"\n')
        f.write('#include "../lifeperm.h"\n')
        f.write('namespace %s {\n\n' % rulestring.replace('-', '_'))

        # We reuse the 2-state isotropic kernel by #including it into an
        # inner namespace:
        f.write('    namespace twostate {\n\n')
        f.write('#include "ma_%s.h"\n\n' % logstring)
        f.write('#include "../isoluts.h"\n')
        f.write('    }\n\n')

        f.write('    bool iterate_var_leaf(uint64_t *inleaves, uint64_t *hleaves, uint64_t *outleaf) {\n\n')
        f.write('        twostate::iterate_var_leaf(inleaves, outleaf);\n')
        f.write('        outleaf[0] &= (~hleaves[3]); outleaf[1] &= (~hleaves[6]); outleaf[2] &= (~hleaves[9]); outleaf[3] &= (~hleaves[12]);\n')
        f.write('        uint64_t outleaf2[4] = {inleaves[3] & (~outleaf[0]), inleaves[6] & (~outleaf[1]), inleaves[9] & (~outleaf[2]), inleaves[12] & (~outleaf[3])};\n')
        postprocess(f, nstates)
        f.write('        return false;\n')
        f.write('    }\n')
        f.write('}\n')



from .parsetable import ReadRuleTable
from .parsetree import ReadRuleTree, SurplusTreeError, FlattenRuleTree, get_symmetries
from .writetree import TransitionsToTree
from .writecode import NodesToCode, MakeStaticTable

import os
import sys
from collections import OrderedDict
from ast import literal_eval

lifelib_dir = os.path.dirname(os.path.abspath(__file__))
while (os.path.basename(lifelib_dir) != 'lifelib'):
    lifelib_dir = os.path.abspath(os.path.join(lifelib_dir, '..'))
rules_dir = os.path.join(lifelib_dir, 'rules')

def table_to_tree(lines):

    n_states, nhood, transitions = ReadRuleTable(lines)
    lines = TransitionsToTree(n_states, nhood, transitions)
    return lines

def minkowski_symmetrise(nhood):

    return set([tuple(((x - y) for (x, y) in zip(t, s))) for t in nhood for s in nhood])

def analyse_tree(segments):

    lines = segments['@TREE']

    n_states, nhood, list_of_nodes = ReadRuleTree(lines)
    syms = get_symmetries(n_states, nhood, list_of_nodes)

    if len(syms) == 12:
        symstring = "HEXAGONAL"
    elif len(syms) == 8:
        symstring = "SQUARE"
    else:
        symstring = "NONE"

    is16bit = (n_states > 256)

    flattened_tree = FlattenRuleTree(n_states, nhood, list_of_nodes)
    segments['@FLATTREE'] = MakeStaticTable(flattened_tree)

    # Determine zone of influence for rule:
    mink = minkowski_symmetrise(nhood)
    chebyshev = max([max(map(abs, t)) for t in mink])
    manhattan = max([sum(map(abs, t)) for t in mink])
    zoi = "9" * (manhattan - chebyshev) + "5" * (chebyshev * 2 - manhattan)

    preamble  = ['/* %d-bit code generated from ruletree */\n' % (16 if is16bit else 8)]
    preamble += ['/* ZOI="%s" */\n' % zoi]
    preamble += ['/* BITS="%d" */\n' % (16 if is16bit else 8)]
    preamble += ['/* NHOOD="%s" */\n' % nhood]
    preamble += ['/* SYMS="%s" */\n' % symstring]

    segments['@PREAMBLE'] = preamble

def create_code(segments):

    lines = []
    template = 0
    nhood = None
    bits = None

    for l in segments['@PREAMBLE']:
        if (nhood is None) and ('NHOOD=' in l):
            nhood = literal_eval(l.split('"')[1])
        if (bits is None) and ('BITS=' in l):
            bits = literal_eval(l.split('"')[1])

    if '@FLATTREE' in segments:
        template += 1
        lines += segments['@FLATTREE']
        lines += NodesToCode(nhood, (bits > 8))

    if '@AIGER' in segments:
        template += 2
        raise NotImplementedError("AIGER is currently unsupported")

    template = ["row_based_approach.h", "aiger_approach.h", "row_and_aiger.h"][template - 1]

    # We have implemented iterate_var_row instead of iterate_var_grid:
    lines.append('#include "%s"\n' % template)

    segments['@CODE'] = lines


def segmentate(rulefile):
    '''
    Split a file into segments demarcated by '@' annotations.
    '''

    segments = OrderedDict()
    seg = '@PREAMBLE'

    for l in rulefile:
        ls = l.strip()
        if ls.startswith('@'):
            seg = ls.split()[0]
        if seg not in segments:
            segments[seg] = []
        segments[seg].append(l)

    return segments

def decapitalise(grulename):
    '''
    Convert a Golly-style rulename with capital letters into a mangled
    lifelib rulename with leading 'x' and sufficient information to
    recover the original Golly rulename.
    '''

    if (grulename == grulename.lower()):
        # Idempotent
        return grulename

    grulename = grulename[0].upper() + grulename[1:]
    rulename = grulename.lower()
    compares = [(x != y) for (x, y) in zip(rulename, grulename)]

    rulename = ('x' if compares[0] else '_') + rulename
    for (i, j) in enumerate(compares):
        if (i > 0) and j:
            rulename = 'x' + str(i) + rulename

    if (rulename[0] != 'x'):
        rulename = 'x' + rulename

    return rulename

def rule2segs(rulefile):
    '''
    Accepts a filename or file-like object as a parameter and
    returns the ordered pair (rulename, ruledata) where ruledata
    is a dictionary mapping segment names to lists of lines.

    This will ensure that the ruledata dictionary contains a
    segment called @CODE containing C/C++ code for the rule.
    '''

    if isinstance(rulefile, str):
        filename = rulefile
        basename = os.path.splitext(os.path.basename(filename))[0]
        with open(filename) as f:
            rulefile = list(f)
        if filename.endswith('.table'):
            rulefile = [('@RULE %s\n' % basename), '@TABLE\n'] + rulefile
        elif filename.endswith('.tree'):
            rulefile = [('@RULE %s\n' % basename), '@TREE\n'] + rulefile
    else:
        rulefile = list(rulefile)
    segments = segmentate(rulefile)

    if '@RULE' not in segments:
        if '@NUTSHELL' not in segments:
            raise RuntimeError("Rule file must contain either a @RULE or @NUTSHELL segment.")

        if (sys.version_info < (3, 6)):
            raise NotImplementedError("Nutshell transpilation requires Python >= 3.6")

        from nutshell.main import transpile

        transpiled = transpile(rulefile).splitlines(True)
        new_segments = segmentate(transpiled)
        segments.update(new_segments)

        if '@RULE' not in segments:
            raise RuntimeError("Nutshell transpilation failed to produce a @RULE segment.")

    grulename = segments['@RULE'][0].strip().split()[1]
    rulename = decapitalise(grulename)

    while '@CODE' not in segments:

        # Apply transformations

        if ('@PREAMBLE' in segments) and (('@FLATTREE' in segments) or ('@AIGER' in segments)):
            create_code(segments)
        elif '@TREE' in segments:
            analyse_tree(segments)
        elif '@TABLE' in segments:
            segments['@TREE'] = table_to_tree(segments['@TABLE'])
        else:
            raise RuntimeError("Rule file must contain a @CODE, @TREE, or @TABLE segment.")

    segments = {k : [x for x in v if (not x.strip().startswith('@'))] for (k, v) in segments.items()}
    return rulename, segments

def rule2files(rulefile, ruledir=rules_dir):

    rulename, segments = rule2segs(rulefile)

    filename = rulename if (rulename[0] == 'x') else ('x' + rulename)

    # Write rule file rulename.h:
    with open(os.path.join(ruledir, filename + '.h'), 'w') as f:
        if '@PREAMBLE' in segments:
            for l in segments['@PREAMBLE']:
                f.write(l)
        f.write('namespace %s {\n' % rulename.replace('-', '_'))
        for l in segments['@CODE']:
            f.write('    ' + l)
        f.write('}\n')

    # Optionally write rulename.colors and rulename.icons:
    for extension in ['colors', 'icons']:
        segname = '@' + extension.upper()
        if segname in segments:
            with open(os.path.join(ruledir, filename + '.' + extension), 'w') as f:
                for l in segments[segname]:
                    f.write(l)

    return rulename



import subprocess

from .eightbit import *
from .eightbit import create_rule as make_header
from .rulefiles import rule2files
from ._caviewer import download


def create_rule(rulestring):
    download()  # Download CAViewer

    preface = os.path.dirname(os.path.abspath(__file__)) + '/bin/CAViewer'
    if rulestring[-2:] in ["v2", "fe", "fc"] or rulestring[-1:] in ["h"]:
        rulestring = rulestring[:-2] + rulestring[-2:].upper()

    # Canonising rulestring
    p = subprocess.Popen(
        preface + " info -r " + rulestring,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    out = p.communicate()

    if out[1].decode("utf-8"):
        raise Exception(out[1].decode("utf-8"))

    # Change from B.../S.../N... to b...s......
    rulestring = out[0].decode("utf-8").split("\n")[0].\
        replace("Rulestring: ", "").replace("/N", "").replace("/S", "s").lower()
    if rulestring[-2:] in ["v2", "fe", "fc"] or rulestring[-1:] in ["h"]:
        rulestring = rulestring[:-2] + rulestring[-2:].upper()

    # Generating rulefile
    p = subprocess.Popen(
        preface + " apgtable -r " + rulestring + " -o " + rulestring.lower() + ".rule",
        stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    out = p.communicate()

    if out[1].decode("utf-8"):
        raise Exception(out[1].decode("utf-8"))

    # Make rulestring lowercase again
    rulestring = rulestring.lower()

    # Convert .rule file into C/C++ code:
    rule2files(rulestring + '.rule')

    # Link code into lifelib:
    make_header(rulestring)




registry = {}

def register_pattern_callback(saveid, pattern):

    registry[saveid] = pattern
    commence_listening()

def rewrite(saveid, rle):

    x = registry[saveid] 
    newpat = x.owner.pattern(rle, x.getrule())
    x &= x.owner.pattern("", x.getrule())
    x += newpat

def commence_listening():

    from IPython.display import display, HTML

    display(HTML('''
    <div>
    <p style="color:green">Press <b>Ctrl+S</b> to save LifeViewer edits.</p>
    <script>
    var declaredListenerYet;
    if (!declaredListenerYet) {
        window.addEventListener('message',function(e) {
            var key = e.message ? 'message' : 'data';
            var data = e[key]
            console.log(data)
            if (data['rle']) {
                var rle = data['rle'].replace(/"/g, ''); // sanitise
                var saveid = data['uuid'].replace(/"/g, '');
                IPython.notebook.kernel.execute('import lifelib; lifelib.registry.rewrite("""' + saveid + '""", """' + rle + '""")');
            }
        },false);
        declaredListenerYet = true;
    }
    </script>
    </div>
    '''))




import os
import lifelib
import unittest
from lifelib.autocompile import sanirule

class TestSanirule(unittest.TestCase):

    def test_goltree(self):

        filename = os.path.join(lifelib.lifelib_dir, 'rules', 'source', 'Life.tree')
        srule = sanirule(filename)
        self.assertEqual(srule, 'b3s23')

    def test_hextree(self):

        filename = os.path.join(lifelib.lifelib_dir, 'rules', 'source', 'Hex-B2omS2.tree')
        srule = sanirule(filename)
        self.assertEqual(srule, 'b2-ps2h')

if __name__ == '__main__':
    unittest.main()




import re
import sys
import os
import subprocess
import time

from . import genera
from ._version import __version__

from .genera import sanirule, genus_to_module, obtain_genus
from .pythlib.lowlevel import regstring

__all__ = ['lifelib_dir', 'compile_rules', 'load_rules', 'reset_tree', 'set_rules',
           'sanirule', 'run_tests', 'add_cygdir', 'install_cygwin']

lifelib_dir = os.path.dirname(os.path.abspath(__file__))
this_python = os.path.abspath(sys.executable)
cygwin_dirs = []

def add_cygdir(*args):

    for a in args:
        cygwin_dirs.append(a)

def get_cygwin_dir():

    cygwin_possibilities = cygwin_dirs + [os.path.join(lifelib_dir, 'cygwin', 'cygwin64'), 'C:\\cygwin64']

    for c in cygwin_possibilities:
        if os.path.exists(c) and os.path.exists(os.path.join(c, 'bin')):
            return c

    raise ValueError("Cygwin directory unknown; please call lifelib.add_cygdir(r'D:\\path\\to\\cygwin64') or lifelib.install_cygwin()")

def install_cygwin(root_directory=None, packages=['python38-numpy', 'gcc-core', 'gcc-g++', 'make', 'git']):
    '''
    For lifelib to run correctly on Windows, there needs to be a 64-bit Cygwin
    installation with g++, Python, and numpy. The user is free to download and
    configure Cygwin herself; this is provided as a speedy alternative which
    includes the prerequisites by default.

    This will eventually result in a directory tree resembling the following:

    lifelib--+--cygwin--+--cygwin-x86_64.exe
             |          |
             |          +--packages
             |          |
             |          +--cygwin64--+--bin-----bash.exe
             |                       |
             +--autocompile.py       +--etc
             |                       |
             +--cygbash.sh           +--usr
             |                       |
           [...]                   [...]

    This has a certain elegance that uninstalling lifelib will remove the
    internal Cygwin installation.
    '''

    try:
        from urllib import urlretrieve
    except ImportError:
        from urllib.request import urlretrieve

    package_directory = os.path.join(lifelib_dir, 'cygwin', 'packages')
    cygwin_setup = os.path.join(lifelib_dir, 'cygwin', 'cygwin-x86_64.exe')

    if root_directory is None:
        root_directory = os.path.join(lifelib_dir, 'cygwin', 'cygwin64')

    print("Preparing to install cygwin64 into %s..." % root_directory)

    if not os.path.exists(package_directory):
        os.makedirs(package_directory)

    # Remove the setup-x86_64 executable if it already exists.
    if os.path.exists(cygwin_setup):
        os.remove(cygwin_setup)

    # By downloading through HTTPS, the setup-x86_64 executable is secure
    # against man-in-the-middle attacks. Beyond that, individual packages
    # are secured by means of digital signatures and SHA-512 hashes; the
    # public-key used to verify those is embedded in setup-x86_64.
    print("Downloading Cygwin setup over HTTPS...")
    urlretrieve("https://cygwin.com/setup-x86_64.exe", cygwin_setup)

    args=[cygwin_setup, '-q', '--no-admin', '-n', '-N', '-d',
            '-R', root_directory, '-l', package_directory,
            '-s', 'http://mirrors.kernel.org/sourceware/cygwin',
            '-P', ','.join(packages)]

    # We are running in Linux, Mac OS X, or already inside Cygwin. In these
    # cases, it does not make sense to install Cygwin.
    if (os.name != 'nt'):
        print("Skipping Cygwin installation since os.name != 'nt'.")
        return

    # Install Cygwin into a subdirectory of lifelib, complete with all of
    # the dependencies used by lifelib.
    print("Installing Cygwin...")
    subprocess.check_call(args)
    verify_installation()

    print("Installation complete.")


def verify_installation():

    print("Verifying installation...")
    subprocess.check_call(get_local_bash() + ['echo', 'SUCCESS'])

    print("Verifying numpy is installed...")
    if (os.name == 'nt'):
        np_version = subprocess.check_output(get_local_bash() + ['python3', '-c', 'import numpy as np; print(np.__version__)'])
    else:
        import numpy as np
        np_version = np.__version__
    np_version = regstring(np_version).strip()
    print("Numpy version: %s" % np_version)
    return np_version


def get_local_bash():

    if (os.name == 'nt'):
        return [os.path.join(get_cygwin_dir(), 'bin', 'bash.exe'), os.path.join(lifelib_dir, 'cygbash.sh')]
    else:
        return ['bash', os.path.join(lifelib_dir, 'cygbash.sh')]


def get_local_python():

    if (os.name == 'nt'):
        return 'python3'
    else:
        return this_python


def get_compiler():

    if (os.name == 'nt'):
        return get_local_bash() + ['g++']
    else:
        return ['g++']


def run_tests():

    print("Warning: if you are in a Jupyter notebook you will not see output until the tests complete.")
    print("Use the terminal to view periodic progress...")
    if (os.name == 'nt'):
        subprocess.check_call(get_local_bash() + ['tests/test_all.sh'])
    else:
        subprocess.check_call(get_local_bash() + ['tests/test_all.sh', this_python])
    print("Completed tests successfully.")


def obtain_arguments(soname, standard="C++11", optimisations=["-march=native", "-O3"],
                    options=["-Wall", "-Wextra", "-fdiagnostics-color=always"], **unused_kwargs):

    so_args = ["-fPIC", "-shared", "-o", soname, "lifelib.cpp"]

    if isinstance(standard, str):
        standard = standard.lower()
        if standard[0] != '-':
            standard = "-std=" + standard
        standard = [standard]

    return standard + optimisations + options + so_args


def compile_rules(*rules, **kwargs):

    if (len(rules) == 0):
        rules = ['b3s23']

    if (len(rules) > 8):
        raise ValueError("Each lifelib session may have up to 8 rules.")

    soname = os.path.join(lifelib_dir, 'pythlib', 'lifelib_%s.so' % '_'.join(rules))

    force_compile = kwargs.get('force_compile', False)
    verify_version = kwargs.get('verify_version', True)

    if force_compile or not os.path.exists(soname):
        recompile = True
    elif verify_version:
        from .pythlib import Session
        local_bash = get_local_bash()
        local_python = kwargs.get('local_python', get_local_python())
        sess = Session(soname, rules=rules, local_bash=local_bash, local_python=local_python)
        compiled_version = sess.lifelib.version
        python_version = __version__
        if compiled_version != python_version:
            sys.stderr.write('Compiled version %s does not match current version %s; recompiling...\n' % (compiled_version, python_version))
            recompile = True
        else:
            recompile = False
    else:
        recompile = False

    if recompile:

        etime = 20.0

        try:
            with open(os.path.join(lifelib_dir, 'compiletime.txt'), 'r') as f:
                for l in f:
                    l = l.strip()
                    if '.' in l:
                        etime = float(l)
        except:
            pass


        pbar = None
        imax = 64

        try:

            cwd = os.path.abspath(os.getcwd())

            print("Generating code for rules %s..." % str(list(rules)))
            generate_code(rules)
            os.chdir(lifelib_dir)

            print("Compiling lifelib shared object (estimated time: %.1f seconds)..." % etime)
            ccargs = get_compiler() + obtain_arguments(soname, **kwargs)
            po = subprocess.Popen(ccargs, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

            i = 0

            try:
                from tqdm.auto import tqdm
                pbar = tqdm(total=100, desc='Estimated compilation progress: ')
                imax = 99
            except:
                sys.stdout.write('[' + (' ' * imax) + ']\r[')

            while po.poll() is None:
                time.sleep(etime / imax)
                if (i < imax):
                    if pbar is None:
                        sys.stdout.write('=')
                        sys.stdout.flush()
                    else:
                        pbar.update(1)
                i += 1

            g_stdout, g_stderr = po.communicate()
            status = po.returncode

            if pbar is None:
                char = '=' if (status == 0) else '!'
                if (i < imax):
                    sys.stdout.write((char * (imax - i)))
                sys.stdout.write(']\n')
                sys.stdout.flush()
            elif (status == 0):
                for j in range(100 - min(imax, i)):
                    pbar.update(1)

            if kwargs.get('compiler_output', (status != 0)):
                sys.stderr.write("Compiler stdout: %s\n" % regstring(g_stdout))
                sys.stderr.write("Compiler stderr: %s\n" % regstring(g_stderr))

            if (status != 0):
                raise RuntimeError("g++ returned nonzero exit code %d" % status)

        finally:

            if pbar is not None:
                pbar.close()
            os.chdir(cwd)

        etime = ((etime * i) / imax) + 0.5

        print('...compilation complete!')

        with open(os.path.join(lifelib_dir, 'compiletime.txt'), 'w') as f:
            f.write(str(etime))

        reset_tree()
            
    return soname


def load_rules(*rules, **kwargs):

    rules = [sanirule(r, drop_history=True) for r in rules]

    soname = compile_rules(*rules, **kwargs)
    from .pythlib import Session
    use_indirection = (os.name == 'nt') or kwargs.get('force_indirect', False)
    local_bash = get_local_bash() if use_indirection else None
    local_python = kwargs.get('local_python', get_local_python())

    # Choose the default number of planes to accommodate every rule
    # for which the lifetree has been compiled, including LifeHistory:
    default_planes = max([genera.rule_property(r, 'bitplanes') for r in rules])
    default_planes = max(4, default_planes)

    return Session(soname, rules=rules, local_bash=local_bash,
                    local_python=local_python, default_planes=default_planes)

def write_grid_iterator(f, rules, families):

    f.write('    void iterate_var_grid(int rule, uint8_t *ingrid, uint8_t *outgrid) {\n')
    f.write('        switch(rule) {\n')
    not_used = True
    for (i, (runsafe, family)) in enumerate(zip(rules, families)):
        r = runsafe.replace('-', '_')

        if (family == 6) or (family == 8):
            not_used = False
            f.write('            case %d :\n' % i)
            f.write('                %s::iterate_var_grid(ingrid, outgrid); break;\n' % r)

    f.write('        }\n')
    if not_used:
        f.write('        (void) ingrid; (void) outgrid;\n')
    else:
        f.write('        #define LIFELIB_FAMILY_SIX 1\n')
    f.write('    }\n\n')

def write_all_iterators(f, hist, rules, families):

    params = 'int n, uint64_t * inleaves'
    params2 = 'uint32_t* d'
    xparams = 'n, inleaves'
    xparams2 = 'd'

    if (hist):
        params += ', uint64_t * hleaves'
        xparams += ', hleaves'
        params2 += ', uint32_t* h'
        xparams2 += ', h'

    if (hist >= 2):
        params += ', uint64_t * jleaves'
        xparams += ', jleaves'
        params2 += ', uint32_t* j'
        xparams2 += ', j'

    params += ', uint64_t * outleaf'
    params2 += ', uint32_t * diffs'
    xparams += ', outleaf'
    xparams2 += ', diffs'

    f.write('    int iterate_var_leaf(int rule, %s) {\n' % params)
    f.write('        switch(rule) {\n')

    not_used = True
    nnot_used = True
    for (i, (runsafe, family)) in enumerate(zip(rules, families)):
        r = runsafe.replace('-', '_')

        if (family == 0):
            not_used = False
            nnot_used = False
            f.write('            case %d :\n' % i)
            f.write('                return %s::iterate_var_leaf(%s);\n' % (r, xparams))
        elif ((family == 2) and (hist == 1)):
            not_used = False
            f.write('            case %d :\n' % i)
            f.write('                return %s::iterate_var_leaf(inleaves, hleaves, outleaf);\n' % r)
        elif ((family == 4) and (hist == 0)):
            not_used = False
            f.write('            case %d :\n' % i)
            f.write('                return %s::iterate_var_leaf(inleaves, outleaf);\n' % r)
    f.write('        }\n')
    if not_used:
        f.write('        (void) %s;\n' % (xparams.replace(',', '; (void)')))
    elif nnot_used:
        f.write('        (void) n;\n')
    f.write('        return -1;\n')
    f.write('    }\n\n')

    def vlife_var(fname, lifeonly):

        f.write('    int %s(int rule, %s) {\n' % (fname, params2))
        f.write('        switch(rule) {\n')

        not_used = True
        for (i, (runsafe, family)) in enumerate(zip(rules, families)):
            r = runsafe.replace('-', '_')
            if ((r == 'b3s23') if lifeonly else (family == 0)):
                not_used = False
                f.write('            case %d :\n' % i)
                f.write('                return %s::%s(%s);\n' % (r, fname, xparams2))
        f.write('        }\n')
        if not_used:
            f.write('        (void) %s;\n' % (xparams2.replace(',', '; (void)')))
        f.write('        return -1;\n')
        f.write('    }\n\n')

    vlife_var('iterate_var_32_28', False)
    f.write('\n\n#ifdef __AVX512F__\n\n')
    vlife_var('iterate_var_48_28', True)
    f.write('\n\n#endif\n\n')


def reset_tree(rule='b3s23', throw_error=False):

    if not isinstance(rule, list):
        rule = [rule]

    try:
        cwd = os.path.abspath(os.getcwd())
        generate_code(rule, clean_before=True)
    except:
        if throw_error:
            raise
        else:
            import warnings
            warnings.warn("Could not reset tree -- check permissions?")
    finally:
        os.chdir(cwd)


def set_rules(*rules):

    reset_tree(list(rules), throw_error=True)


def generate_code(rules, clean_before=False):

    logic_directory = 'avxlife/lifelogic'

    # Ensure modules are importable:
    modules = [genus_to_module(obtain_genus(rulestring)) for rulestring in rules]

    os.chdir(lifelib_dir)

    if clean_before and os.path.exists(logic_directory):
        import shutil
        shutil.rmtree(logic_directory)

    if not os.path.exists(logic_directory):
        os.makedirs(logic_directory)
    os.chdir(logic_directory)

    # Create the source code for the rules:
    for r in rules:

        genera.create_rule(r)

    # Determine rule families from genera:
    try:
        families = [genera.rule_property(r, 'family') for r in rules]
    except ImportError:
        print("ImportError; current directory == %s" % os.getcwd())
        raise

    # Obtain integers describing valid mantissae:
    def mant2int(m):
        if hasattr(m, '__iter__'):
            m = sum([(1 << x) for x in range(9) if x in m])
        return (m | 1)
    mantissae = [mant2int(genera.rule_property(r, 'mantissa')) for r in rules]

    # Write the main entry point:
    with open('iterators_all.h', 'w') as f:

        f.write('#pragma once\n')
        f.write('#include <string>\n')
        f.write('#include <vector>\n')

        #include the generated code for each rule:
        for rulestring in rules:
            f.write('#include "iterators_%s.h"\n' % rulestring)

        f.write('namespace apg {\n\n')

        # Function to convert rulestring to rule integer:

        for prop in ['zoi', 'syms']:

            not_used = True
            f.write('    std::string get_%s(std::string rule) {\n' % prop)
            for (fam, r) in zip(families, rules):
                if (fam >= 6):
                    f.write('        if (rule == "%s") { return "%s"; }\n' % (r, genera.rule_property(r, prop)))
                    not_used = False
            if not_used:
                f.write('        (void) rule;\n')
            f.write('        return "99";\n')
            f.write('    }\n\n')

        # Function to convert rulestring to rule integer:
        f.write('    int rule2int(std::string rule) {\n')
        for (i, r) in enumerate(rules):
            f.write('        if (rule == "%s") { return %d; }\n' % (r, i))
        f.write('        return -1;\n')
        f.write('    }\n\n')

        # Function to return all rulestrings:
        f.write('    std::vector<std::string> get_all_rules() {\n')
        f.write('        std::vector<std::string> x;\n')
        for r in rules:
            f.write('        x.emplace_back("%s");\n' % r)
        f.write('        return x;\n')
        f.write('    }\n\n')

        # Function to obtain family code from rule integer:
        f.write('    int uli_get_family(int rule) {\n')
        f.write('        switch (rule) {\n')
        for x in enumerate(families):
            f.write('            case %d : return %d;\n' % x)
        f.write('        }\n')
        f.write('        return 0;\n')
        f.write('    }\n\n')

        # Function to obtain mantissae from rule integer:
        f.write('    uint64_t uli_valid_mantissa(int rule) {\n')
        f.write('        switch (rule) {\n')
        for x in enumerate(mantissae):
            f.write('            case %d : return %d;\n' % x)
        f.write('        }\n')
        f.write('        return 3;\n')
        f.write('    }\n\n')

        for hist in range(3):
            write_all_iterators(f, hist, rules, families)

        write_grid_iterator(f, rules, families)

        f.write('}\n')

    os.chdir(lifelib_dir)

